from enum import StrEnum
from pathlib import Path
from typing import Sequence, Iterable

from pydantic import BaseModel

from ..utils import load_lines, write_lines


def format_cookies(cookies):
    """
    # Фикс непонятно чего..
    """
    for cookie in cookies:
        if cookie.get('sameSite') == 'no_restriction':
            cookie['sameSite'] = 'None'


class GoogleAccountStatus(StrEnum):
    UNKNOWN = "UNKNOWN"
    BAD_COOKIES = "BAD_COOKIES"
    RECOVERY_EMAIL_REQUIRED = "RECOVERY_EMAIL_REQUIRED"
    RECOVERY_REQUIRED = "RECOVERY_REQUIRED"
    CAPTCHA_REQUIRED = "CAPTCHA_REQUIRED"
    PHONE_VERIFICATION_REQUIRED = "PHONE_VERIFICATION_REQUIRED"
    BANNED = "BANNED"
    GOOD = "GOOD"

    def __str__(self):
        return self.value


class GoogleAccount(BaseModel):
    email:          str
    password:       str
    recovery_email: str | None = None
    # TODO Валидировать cookies
    cookies:        list | None = None
    status: GoogleAccountStatus = GoogleAccountStatus.UNKNOWN

    def __init__(
            self,
            email: str,
            password: str,
            *,
            recovery_email: str = None,
            cookies: list = None
    ):
        super().__init__(
            email=email,
            password=password,
            recovery_email=recovery_email,
            cookies=cookies,
        )

    @property
    def short_password(self) -> str:
        start = self.password[:2]
        end = self.password[-2:]
        return f"{start}**{end}"

    def __repr__(self):
        return f"{self.__class__.__name__}(email={self.email})"

    def __str__(self):
        return self.email

    @classmethod
    def from_file(
            cls,
            filepath: Path | str,
            *,
            separator: str = ":",
            fields: Sequence[str] = ("email", "password", "recovery_email"),
    ):
        """
        :param filepath: Путь до файла с данными об аккаунтах.
        :param separator: Разделитель между данными в строке.
        :param fields: Кортеж, содержащий имена полей в порядке их появления в строке.
         Должно содержать как минимум два поля - email и password: `("email", )`
        :return: Список аккаунтов.
        """
        accounts = []
        for line in load_lines(filepath):
            data = dict(zip(fields, line.split(separator)))
            data.update({key: None for key in data if not data[key]})
            accounts.append(cls(**data))

        return accounts

    @classmethod
    def to_file(
            cls,
            filepath: Path | str,
            accounts: Iterable["GoogleAccount"],
            *,
            separator: str = ";",
            fields: Sequence[str] = ("email", "password", "recovery_email"),
    ):
        lines = []
        for account in accounts:
            account_data = []
            for field_name in fields:
                field = getattr(account, field_name)
                field = field if field is not None else ""
                account_data.append(field)
            lines.append(separator.join(account_data))
        write_lines(filepath, lines)



import warnings
from . import utils, base, twitter, discord, googleapis, google


# HACK: Ignore event loop warnings from curl_cffi
warnings.filterwarnings('ignore', module='curl_cffi')


__all__ = [
    "utils",
    "base",
    "twitter",
    "discord",
    "googleapis",
    "google",
]



from contextlib import asynccontextmanager

from playwright.async_api import async_playwright, Browser, Playwright
from better_proxy import Proxy


class BasePlaywrightBrowser:
    """
    Базовый асинхронный Playwright браузер:
        - Принимает прокси в формате URL и better-proxy.
        - Устанавливает таймаут в 10 сек по умолчанию.
    """
    proxy: Proxy | None

    def __init__(
            self,
            *,
            default_timeout: int = 10_000,
            proxy: str | Proxy = None,  # TODO Принимать в Playwright формате тоже
            **launch_kwargs
    ):
        self._playwright: Playwright | None = None
        self._browser: Browser | None = None
        self.proxy = Proxy.from_str(proxy) if proxy else None
        self.launch_kwargs = launch_kwargs
        self.default_timeout = default_timeout

    async def create_browser(self):
        self._playwright = await async_playwright().start()
        proxy = self.proxy.as_playwright_proxy if self.proxy else None
        self._browser = await self._playwright.firefox.launch(proxy=proxy, **self.launch_kwargs)

    async def close_browser(self):
        await self._browser.close()
        await self._playwright.stop()

    async def __aenter__(self):
        await self.create_browser()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close_browser()

    @asynccontextmanager
    async def new_context(
            self,
            *,
            proxy: str | Proxy = None,  # TODO Принимать в Playwright формате тоже
            **context_kwargs,
    ):
        proxy = Proxy.from_str(proxy).as_playwright_proxy if proxy else None
        context = await self._browser.new_context(proxy=proxy, **context_kwargs)
        context.set_default_timeout(self.default_timeout)
        # thx @ruslanwuwei
        await context.add_init_script("delete Object.getPrototypeOf(navigator).webdriver")
        yield context
        await context.close()



import re
from typing import Literal

from yarl import URL
from playwright.async_api import BrowserContext, Request, TimeoutError as PlaywrightTimeoutError, Page
from playwright_stealth import stealth_async

from .errors import (
    CaptchaRequired,
    FailedToOAuth2,
    FailedToLogin,
    RecoveryRequired,
    RecoveryEmailRequired,
    PhoneVerificationRequired,
)
from .account import GoogleAccount, GoogleAccountStatus
from .utils import check_cookies
from ..sms import SmshubClient
from ..sms.errors import SmsServiceError


PromptType = Literal["consent", "select_account"] | None


def are_valid_google_cookies(cookies: list[dict]) -> bool:
    """
    SID и HSID: Эти cookie содержат цифровые подписи и информацию о последнем входе в систему.
    SSID, APISID, SAPISID: Также содержат информацию об аутентификации и используются в различных сервисах Google для поддержания сессии пользователя.
    __Secure*: Хотя они в первую очередь используются для рекламных целей, они также связаны с твоей учетной записью и могут содержать важную информацию о сессии.
    NID: Используется для хранения настроек пользователя и может содержать информацию, упрощающую доступ к аккаунту.
    GAPS: Этот cookie используется для аутентификации в различных приложениях Google и может содержать важные данные аутентификации.

    Все cookie, содержащие информацию об аутентификации:
        SID
        HSID
        SSID
        APISID
        SAPISID
        OTZ
        NID
        OSID
        LSID
        SIDCC
        ACCOUNT_CHOOSER
        __Secure-1PSIDTS
        __Secure-3PSIDTS
        __Secure-1PSID
        __Secure-3PSID
        __Secure-1PAPISID
        __Secure-3PAPISID
        __Secure-1PSIDCC
        __Secure-3PSIDCC
        __Secure-OSID
        __Host-1PLSID
        __Host-3PLSID
        __Host-GAPS
    """
    return check_cookies(cookies, {"SID", "HSID"})


class GooglePlaywrightBrowserContext:
    """
    Для решения капчи используется capsolver. Для SMS - smshub.
    В будущих версиях будет добавлено гораздо больше вариантов решения капчи и сервисов для приема SMS.
    """

    # Common
    _RECAPTCHA_IFRAME_XPATH = '//iframe[@title="reCAPTCHA"]'
    _RECAPTCHA_CHECKBOX_CHECKED_XPATH = '//span[contains(@class, "recaptcha-checkbox-checked")]'
    _LEFT_BUTTON_XPATH = '//div[@jsname="QkNstf"]/div/div/button'  # Not now, Try another way
    _RIGHT_BUTTON_XPATH = '//div[@jsname="Njthtb"]/div/button'  # Continue, Send, Next
    # _ERRORS_COUNT_XPATH = '//div[@jsname="B34EJ"]'

    # Logining
    _EMAIL_FIELD_XPATH = '//*[@id="identifierId"]'
    _EMAIL_CONFIRMATION_BUTTON_XPATH = '//*[@id="identifierNext"]/div/button'
    _PASSWORD_FIELD_XPATH = '//*[@id="password"]/div[1]/div/div[1]/input'
    _PASSWORD_CONFIRMATION_BUTTON_XPATH = '//*[@id="passwordNext"]/div/button'
    _RECOVERY_EMAIL_BUTTON_XPATH = '//div[@data-challengeid="5"]'
    _RECOVERY_EMAIL_FIELD_XPATH = '//input[@id="knowledge-preregistered-email-response"]'
    _RECOVERY_BUTTON_XPATH = '//div[@id="accountRecoveryButton"]/div/div/a'

    # Logining: patterns
    _RECOVERY_REQUIRED_URL_PATTERN = re.compile(r"https://accounts\.google\.com/v3/signin/rejected.*")
    _PASSKEY_URL_PATTERN = re.compile(r"https://accounts\.google\.com/signin/v2/passkeyenrollment.*")
    _MY_ACCOUNT_URL_PATTERN = re.compile(r"https://myaccount\.google\.com.*")
    _GDS_URL_PATTERN = re.compile(r"https://gds\.google\.com.*")
    _LOGGED_IN_URL_PATTERNS = (_MY_ACCOUNT_URL_PATTERN, _GDS_URL_PATTERN)

    # OAuth2
    _CONTINUE_BUTTON_XPATH = '//div[@jsname="uRHG6"]/div/button'
    _ACCOUNT_BUTTON_XPATH = '//*[@data-identifier="{email}"]'

    # Phone
    _PHONE_NUMBER_INPUT_FIELD_XPATH = '//input[@id="deviceAddress"]'
    _COUNTRY_SELECT_MENU_XPATH = '//select[@id="countryList"]'
    _CODE_INPUT_FIELD = '//*[@id="smsUserPin"]'
    _NEXT_BUTTON_XPATH = '//*[@id="next-button"]'
    _ERROR_SPAN_XPATH = '//span[@id="error"]'

    def __init__(
            self,
            context: BrowserContext,
            account: GoogleAccount,
            *,
            stealth: bool = False,
            time_to_wait: int = 10_000,
            wait_for_captcha_solving: bool = True,
            time_to_solve_captcha: int = 30_000,  # None: default timeout, 0: unlimited
            # capsolver_api_key: str = None,
            smshub_api_key: str = None,
            max_attempts_to_verify_phone_number: int = 5,
    ):
        self._context = context
        self.account = account
        self.stealth = stealth
        self.time_to_wait = time_to_wait
        self.wait_for_captcha_solving = wait_for_captcha_solving
        self.time_to_solve_captcha = time_to_solve_captcha
        # self.capsolver_api_key = capsolver_api_key
        self.smshub_api_key = smshub_api_key
        self.max_attempts_to_verify_phone_number = max_attempts_to_verify_phone_number

        self._logged_in: bool = False
        self._needs_recovery_email: bool = False

    def _account_button_xpath(self) -> str:
        return self._ACCOUNT_BUTTON_XPATH.format(email=self.account.email.lower())

    async def _new_page(self):
        page = await self._context.new_page()
        if self.stealth: await stealth_async(page)
        return page

    async def _location_href(self, page) -> str:
        return await page.evaluate("location.href")

    def logged_in(self) -> bool:
        return self._logged_in

    async def _type_password_with_confirmation(self, page: Page):
        password_input_field = page.locator(self._PASSWORD_FIELD_XPATH)
        password_confirmation_button = page.locator(self._PASSWORD_CONFIRMATION_BUTTON_XPATH)

        # Иногда пароль не печатается с первого раза по неизвестной причине
        while True:
            await password_input_field.type(self.account.password)
            password_input_field_value = await password_input_field.input_value()
            if password_input_field_value: break

        await password_confirmation_button.click()

    async def _check_recovery_email_verification(self, page: Page):
        """
        Если в аккаунт с этого IP не входили ранее, то просит ввести recovery_email.
        Также может попросить recovery_email после прохождения капчи, что будет означать, что аккаунт мертв.

        https://accounts.google.com/v3/signin/challenge/selection
        """
        recovery_email_button = page.locator(self._RECOVERY_EMAIL_BUTTON_XPATH)
        try:
            await recovery_email_button.wait_for(timeout=self.time_to_wait)
            self._needs_recovery_email = True
        except PlaywrightTimeoutError:
            pass

        if self._needs_recovery_email:
            if not self.account.recovery_email:
                self.account.status = GoogleAccountStatus.RECOVERY_EMAIL_REQUIRED
                raise RecoveryEmailRequired(f"Failed to login Google account: recovery email required.")

            await recovery_email_button.click()
            await page.wait_for_load_state("load")
            await page.locator(self._RECOVERY_EMAIL_FIELD_XPATH).type(self.account.recovery_email)
            await page.locator(self._RIGHT_BUTTON_XPATH).click()
            self._needs_recovery_email = False

        await page.wait_for_load_state("load")

        try:
            await page.locator(self._RECOVERY_BUTTON_XPATH).wait_for(timeout=self.time_to_wait)
            self.account.status = GoogleAccountStatus.RECOVERY_REQUIRED
            raise RecoveryRequired("Failed to login Google account."
                                   " Google: We noticed unusual activity in your Google Account."
                                   " To keep your account safe, you were signed out."
                                   " To continue, you’ll need to verify it’s you.")
        except PlaywrightTimeoutError:
            pass

        await page.wait_for_load_state("load")

        # https://accounts.google.com/v3/signin/challenge
        # To help keep your account safe, Google wants to make sure it’s really you trying to sign in
        # Open a browser on a phone or computer where you’re already signed in and go to:
        # https://g.co/verifyaccount
        if (await page.locator(self._LEFT_BUTTON_XPATH).count() > 0
                and await page.locator(self._RIGHT_BUTTON_XPATH).count() == 0):
            raise RecoveryRequired("Failed to login Google account."
                                   " Google: To help keep your account safe, Google wants to make sure it’s really you trying to sign in"
                                   " Open a browser on a phone or computer where you’re already signed in and go to:"
                                   " https://g.co/verifyaccount")

    async def _check_phone_verification(self, page: Page):
        """Привязка номера"""
        country_select_menu = page.locator(self._COUNTRY_SELECT_MENU_XPATH)
        phone_number_input_field = page.locator(self._PHONE_NUMBER_INPUT_FIELD_XPATH)
        code_input_field = page.locator(self._CODE_INPUT_FIELD)
        next_button = page.locator(self._NEXT_BUTTON_XPATH)
        error_span = page.locator(self._ERROR_SPAN_XPATH)

        if "https://accounts.google.com/speedbump/idvreenable" in await self._location_href(page):
            self.account.status = GoogleAccountStatus.PHONE_VERIFICATION_REQUIRED

            if not self.smshub_api_key:
                raise ValueError("No smshub API key")

            try:
                async with SmshubClient(self.smshub_api_key) as smshub:
                    attempt = 0
                    while attempt <= self.max_attempts_to_verify_phone_number:
                        attempt += 1
                        # TODO Позволять пользователю выбирать страну
                        await country_select_menu.select_option(value='ID')  # Indonesia
                        id, number = await smshub.request_number("go", 6)  # Indonesia
                        print(f"({attempt}) {id}: {number}")
                        await phone_number_input_field.type(number)
                        await next_button.click()
                        await page.wait_for_load_state("networkidle")
                        if not await error_span.count():
                            break
                        else:
                            print(await error_span.inner_text())
                            # Если номер невалидный, появляется сообщение с ошибкой
                            await smshub.cancel_activation(id)

                    if attempt <= self.max_attempts_to_verify_phone_number:
                        code = await smshub.wait_for_code(id)
                        await code_input_field.type(code)
                        await next_button.click()
                        return
            except (PlaywrightTimeoutError, TimeoutError, SmsServiceError):
                pass
            raise PhoneVerificationRequired("Phone verification required.")

    async def _check_captcha_and_type_password(self, page: Page, login: bool = True):
        recaptcha_iframe = page.locator(self._RECAPTCHA_IFRAME_XPATH)
        await page.wait_for_load_state("networkidle")
        try:
            await recaptcha_iframe.wait_for(timeout=self.time_to_wait)
            self.account.status = GoogleAccountStatus.CAPTCHA_REQUIRED
        except PlaywrightTimeoutError:
            pass

        if self.account.status == GoogleAccountStatus.CAPTCHA_REQUIRED:
            if self.wait_for_captcha_solving:
                try:
                    recaptcha_frame_name = await recaptcha_iframe.get_attribute("name")
                    recaptcha = page.frame(name=recaptcha_frame_name)
                    print("жду решения рекапчи")
                    await recaptcha.locator(self._RECAPTCHA_CHECKBOX_CHECKED_XPATH).wait_for(
                        timeout=self.time_to_solve_captcha)
                    await page.locator(self._RIGHT_BUTTON_XPATH).click()
                    await self._type_password_with_confirmation(page)
                    await self._check_recovery_email_verification(page)
                    await self._check_phone_verification(page)
                    return
                except PlaywrightTimeoutError:
                    pass
            raise CaptchaRequired("Failed to login Google account: captcha required.")
        elif login:
            await self._type_password_with_confirmation(page)
            await self._check_recovery_email_verification(page)

    async def login(self):
        if self.account.cookies:
            if not are_valid_google_cookies(self.account.cookies):
                self.account.status = GoogleAccountStatus.BAD_COOKIES
                return

            await self._context.add_cookies(self.account.cookies)
            self.account.status = GoogleAccountStatus.GOOD
            self._logged_in = True
            return

        page = await self._new_page()
        try:
            await page.goto("https://accounts.google.com/ServiceLogin")
            await page.locator(self._EMAIL_FIELD_XPATH).type(self.account.email)
            await page.locator(self._EMAIL_CONFIRMATION_BUTTON_XPATH).click()
            await self._check_captcha_and_type_password(page)

            # Иногда просит установить passkey
            if self._PASSKEY_URL_PATTERN.search(page.url):
                # Not now button
                await page.locator(self._LEFT_BUTTON_XPATH).click(timeout=self.time_to_wait)

            await page.wait_for_load_state("load")

            cookies = None
            for _ in range(5):
                await page.wait_for_timeout(self.time_to_wait / 5)
                if any(url_pattern.search(page.url) for url_pattern in self._LOGGED_IN_URL_PATTERNS):
                    cookies = await self._context.cookies()
                    self._logged_in = are_valid_google_cookies(cookies)
                    break

            await page.close()

            if self._logged_in:
                self.account.status = GoogleAccountStatus.GOOD
                self.account.cookies = cookies
            else:
                self.account.status = GoogleAccountStatus.UNKNOWN
                raise FailedToLogin("Failed to login Google account: failed to catch auth cookies.")

        except PlaywrightTimeoutError:
            await page.close()
            raise FailedToLogin("Failed to login Google account: unexpected TimeoutError.")

    async def oauth2(
            self,
            *,
            client_id: str,
            redirect_uri: str,
            scope: str,
            gsiwebsdk: int = 3,
            access_type: str = "offline",
            response_type: str = "code",
            prompt: PromptType = None,
            include_granted_scopes: bool | str = True,
            enable_granular_consent: bool | str = True,
    ) -> tuple[str | None, str | None]:
        """
        Найдите подобную ссылку: `https://accounts.google.com/o/oauth2/v2/auth?client_id=...`
        Передайте параметры ссылки (после знака вопроса (?)) в метод oauth2
        Метод вернет oauth_code и redirect_url (также содержится в redirect_url)
        :return: oauth_code, redirect_url
        """
        if not self._logged_in:
            await self.login()

        oauth_url = "https://accounts.google.com/o/oauth2/v2/auth"
        params = {
            "client_id": client_id,
            "redirect_uri": redirect_uri,
            "scope": scope,
            "gsiwebsdk": gsiwebsdk,
            "access_type": access_type,
            "response_type": response_type,
            "include_granted_scopes": str(include_granted_scopes).lower(),
            "enable_granular_consent": str(enable_granular_consent).lower(),
        }
        if prompt: params["prompt"] = prompt
        oauth_url = str(URL(oauth_url).with_query(params))
        page = await self._new_page()

        oauth_code = None
        redirect_url = None

        async def request_handler(request: Request):
            nonlocal oauth_code
            nonlocal redirect_url

            # Поимка oauth_code и redirect_url основана на знании того, что гугл делает такой редирект:
            # https://developers.google.com/identity/protocols/oauth2/javascript-implicit-flow#redirecting
            if request.url.startswith(redirect_uri):
                redirect_url = URL(request.url)
                oauth_code = redirect_url.query.get(response_type)

        page.on("request", request_handler)

        try:
            await page.goto(oauth_url)
            # TODO Поведение страницы может отличаться, если значение prompt != "consent"
            await page.wait_for_timeout(self.time_to_wait)
            await page.locator(self._account_button_xpath()).click()
            await self._check_captcha_and_type_password(page, login=False)
            try:
                await page.locator(self._CONTINUE_BUTTON_XPATH).click(timeout=self.time_to_wait)
            except PlaywrightTimeoutError:
                pass
            await page.wait_for_timeout(self.time_to_wait)
        except PlaywrightTimeoutError:
            await page.close()
            raise FailedToOAuth2("Failed to OAuth2 Google account: unexpected TimeoutError.")

        await page.close()

        if not oauth_code:
            raise FailedToOAuth2("Failed to OAuth2 Google account: Failed to catch oauth code.")

        return oauth_code, str(redirect_url)



import asyncio
import base64
from typing import Any, Literal
from time import time

from curl_cffi import requests
from yarl import URL

from better_automation.twitter.errors import (
    HTTPException,
    BadRequest,
    Unauthorized,
    Forbidden,
    NotFound,
    RateLimited,
    TwitterServerError,
    BadTwitterAccount,
    BadToken,
    Locked,
    Suspended,
)
from ..utils import to_json
from ..base import BaseClient
from .account import TwitterAccount, TwitterAccountStatus
from .models import TwitterUserData
from .utils import remove_at_sign, parse_oauth_html


class TwitterClient(BaseClient):
    DEFAULT_HEADERS = {
        'authority': 'twitter.com',
        'origin': 'https://twitter.com',
        'authorization': 'Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs%3D1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA',
        'x-twitter-active-user': 'yes',
        'x-twitter-auth-type': 'OAuth2Session',
        'x-twitter-client-language': 'en',
    }
    GRAPHQL_URL = 'https://twitter.com/i/api/graphql'
    ACTION_TO_QUERY_ID = {
        'CreateRetweet': "ojPdsZsimiJrUGLR1sjUtA",
        'FavoriteTweet': "lI07N6Otwv1PhnEgXILM7A",
        'UnfavoriteTweet': "ZYKSe-w7KEslx3JhSIk5LA",
        'CreateTweet': "SoVnbfCycZ7fERGCwpZkYA",
        'TweetResultByRestId': "V3vfsYzNEyD9tsf4xoFRgw",
        'ModerateTweet': "p'jF:GVqCjTcZol0xcBJjw",
        'DeleteTweet': "VaenaVgh5q5ih7kvyVjgtg",
        'UserTweets': "Uuw5X2n3tuGE_SatnXUqLA",
        'TweetDetail': 'VWFGPVAGkZMGRKGe3GFFnA',
        'ProfileSpotlightsQuery': '9zwVLJ48lmVUk8u_Gh9DmA',
        'Following': 't-BPOrMIduGUJWO_LxcvNQ',
        'Followers': '3yX7xr2hKjcZYnXt6cU6lQ',
        'UserByScreenName': 'G3KGOASz96M-Qu0nwmGXNg',
    }

    @classmethod
    def _action_to_url(cls, action: str) -> tuple[str, str]:
        """
        :return: URL and Query ID
        """
        query_id = cls.ACTION_TO_QUERY_ID[action]
        url = f"{cls.GRAPHQL_URL}/{query_id}/{action}"
        return url, query_id

    def __init__(
            self,
            account: TwitterAccount,
            *,
            wait_on_rate_limit: bool = True,
            **session_kwargs,
    ):
        super().__init__(**session_kwargs)
        self.account = account
        self.wait_on_rate_limit = wait_on_rate_limit

    async def _authenticated_request(self, method, url, **kwargs) -> requests.Response:
        cookies = kwargs["cookies"] = kwargs.get("cookies") or {}
        headers = kwargs["headers"] = kwargs.get("headers") or {}

        cookies["auth_token"] = self.account.auth_token
        if self.account.ct0:
            cookies["ct0"] = self.account.ct0
            headers["x-csrf-token"] = self.account.ct0

        try:
            response = await self.session.request(method, url, **kwargs)
        except requests.errors.RequestsError as exc:
            if exc.code == 35:
                msg = "The IP address may have been blocked by Twitter. Blocked countries: Russia. " + str(exc)
                raise requests.errors.RequestsError(msg, 35, exc.response)
            raise

        if response.status_code == 429:
            if self.wait_on_rate_limit:
                reset_time = int(response.headers["x-rate-limit-reset"])
                sleep_time = reset_time - int(time()) + 1
                if sleep_time > 0:
                    await asyncio.sleep(sleep_time)
                return await self._authenticated_request(method, url, **kwargs)
            raise RateLimited(response, None)

        return response

    async def request(
            self,
            method,
            url,
            **kwargs,
    ) -> tuple[requests.Response, Any]:
        response = await self._authenticated_request(method, url, **kwargs)
        response_json = response.json()

        if response.status_code == 400:
            raise BadRequest(response, response_json)

        if response.status_code == 401:
            exc = Unauthorized(response, response_json)

            if 32 in exc.api_codes:
                self.account.status = TwitterAccountStatus.BAD_TOKEN
                raise BadToken(self.account)

            raise exc

        if response.status_code == 403:
            exc = Forbidden(response, response_json)

            if 353 in exc.api_codes and "ct0" in response.cookies:
                self.account.ct0 = response.cookies["ct0"]
                return await self.request(method, url, **kwargs)

            if 64 in exc.api_codes:
                self.account.status = TwitterAccountStatus.SUSPENDED
                raise Suspended(self.account)

            if 326 in exc.api_codes:
                self.account.status = TwitterAccountStatus.LOCKED
                raise Locked(self.account)

            raise exc

        if response.status_code == 404:
            raise NotFound(response, response_json)

        if response.status_code >= 500:
            raise TwitterServerError(response, response_json)

        if not 200 <= response.status_code < 300:
            raise HTTPException(response, response_json)

        if "errors" in response_json:
            exc = HTTPException(response, response_json)

            if 141 in exc.api_codes:
                self.account.status = TwitterAccountStatus.SUSPENDED
                raise Suspended(self.account)

            if 326 in exc.api_codes:
                self.account.status = TwitterAccountStatus.LOCKED
                raise Locked(self.account)

            raise exc

        self.account.status = TwitterAccountStatus.GOOD
        return response, response_json

    async def _request_oauth_2_auth_code(
            self,
            client_id: str,
            code_challenge: str,
            state: str,
            redirect_uri: str,
            code_challenge_method: str,
            scope: str,
            response_type: str,
    ) -> str:
        url = "https://twitter.com/i/api/2/oauth2/authorize"
        querystring = {
            "client_id": client_id,
            "code_challenge": code_challenge,
            "code_challenge_method": code_challenge_method,
            "state": state,
            "scope": scope,
            "response_type": response_type,
            "redirect_uri": redirect_uri,
        }
        response, response_json = await self.request("GET", url, params=querystring)
        auth_code = response_json["auth_code"]
        return auth_code

    async def _confirm_oauth_2(self, auth_code: str):
        data = {
            'approval': 'true',
            'code': auth_code,
        }
        headers = {'content-type': 'application/x-www-form-urlencoded'}
        await self.request("POST", 'https://twitter.com/i/api/2/oauth2/authorize', headers=headers, data=data)

    async def oauth_2(
            self,
            client_id: str,
            code_challenge: str,
            state: str,
            redirect_uri: str,
            code_challenge_method: str,
            scope: str,
            response_type: str,
    ):
        """
        Запрашивает код авторизации для OAuth 2.0 авторизации.

        Привязка (бинд, линк) приложения.

        :param client_id: Идентификатор клиента, используемый для OAuth.
        :param code_challenge: Код-вызов, используемый для PKCE (Proof Key for Code Exchange).
        :param state: Уникальная строка состояния для предотвращения CSRF-атак.
        :param redirect_uri: URI перенаправления, на который будет отправлен ответ.
        :param code_challenge_method: Метод, используемый для преобразования code_verifier в code_challenge.
        :param scope: Строка областей доступа, запрашиваемых у пользователя.
        :param response_type: Тип ответа, который ожидается от сервера авторизации.
        :return: Код авторизации (привязки).
        """
        auth_code = await self._request_oauth_2_auth_code(
            client_id, code_challenge, state, redirect_uri, code_challenge_method, scope, response_type,
        )
        await self._confirm_oauth_2(auth_code)
        return auth_code

    async def _request_oauth_html(self, oauth_token: str, **oauth_params) -> str:
        """

        :return: html страница привязки приложения (аутентификации) старого типа.
        """
        url = "https://api.twitter.com/oauth/authenticate"
        oauth_params["oauth_token"] = oauth_token
        response = await self._authenticated_request("GET", url, params=oauth_params)

        if response.status_code == 403:
            raise ValueError("The request token (oauth_token) for this page is invalid."
                             " It may have already been used, or expired because it is too old.")

        return response.text

    async def _confirm_oauth(self, oauth_token: str, authenticity_token: str, redirect_after_login_url: str):
        url = "https://api.twitter.com/oauth/authorize"
        params = {
            "redirect_after_login": redirect_after_login_url,
            "authenticity_token": authenticity_token,
            "oauth_token": oauth_token,
        }
        response = await self._authenticated_request("POST", url, data=params)
        return response.text

    async def oauth(self, oauth_token: str, **oauth_params) -> tuple[str, str]:
        """
        :return: redirect_url, authenticity_token
        """
        oauth_data = parse_oauth_html(await self._request_oauth_html(oauth_token, **oauth_params))
        if "redirect_after_login_url" in oauth_data:
            oauth_data = parse_oauth_html(await self._confirm_oauth(
                oauth_token, oauth_data["authenticity_token"], oauth_data["redirect_after_login_url"]))
        return oauth_data["redirect_url"], oauth_data["authenticity_token"]

    async def request_username(self):
        url = "https://twitter.com/i/api/1.1/account/settings.json"
        response, response_json = await self.request("POST", url)
        self.account.username = response_json["screen_name"]

    async def _request_user_data(self, username: str) -> TwitterUserData:
        url, query_id = self._action_to_url("UserByScreenName")
        username = remove_at_sign(username)
        variables = {
            "screen_name": username,
            "withSafetyModeUserFields": True,
        }
        features = {
            "hidden_profile_likes_enabled": True,
            "hidden_profile_subscriptions_enabled": True,
            "responsive_web_graphql_exclude_directive_enabled": True,
            "verified_phone_label_enabled": False,
            "subscriptions_verification_info_is_identity_verified_enabled": True,
            "subscriptions_verification_info_verified_since_enabled": True,
            "highlights_tweets_tab_ui_enabled": True,
            "creator_subscriptions_tweet_preview_api_enabled": True,
            "responsive_web_graphql_skip_user_profile_image_extensions_enabled": False,
            "responsive_web_graphql_timeline_navigation_enabled": True,
        }
        field_toggles = {
            "withAuxiliaryUserLabels": False,
        }
        params = {
            "variables": to_json(variables),
            "features": to_json(features),
            "fieldToggles": to_json(field_toggles),
        }
        response, response_json = await self.request("GET", url, params=params)
        user_data = TwitterUserData.from_raw_user_data(response_json["data"]["user"]["result"])

        if self.account.username == user_data.username:
            self.account.id = user_data.id
            self.account.name = user_data.name

        return user_data

    async def request_user_data(self, username: str = None) -> TwitterUserData:
        if username:
            return await self._request_user_data(username)
        else:
            if not self.account.username:
                await self.request_username()
            return await self._request_user_data(self.account.username)

    async def upload_image(self, image: bytes) -> int:
        """
        Upload image as bytes.

        :return: Media ID
        """
        url = "https://upload.twitter.com/1.1/media/upload.json"

        data = {"media_data": base64.b64encode(image)}
        response, response_json = await self.request("POST", url, data=data)
        media_id = response_json["media_id"]
        return media_id

    async def _follow_action(self, action: str, user_id: int | str) -> bool:
        url = f"https://twitter.com/i/api/1.1/friendships/{action}.json"
        params = {
            'include_profile_interstitial_type': '1',
            'include_blocking': '1',
            'include_blocked_by': '1',
            'include_followed_by': '1',
            'include_want_retweets': '1',
            'include_mute_edge': '1',
            'include_can_dm': '1',
            'include_can_media_tag': '1',
            'include_ext_has_nft_avatar': '1',
            'include_ext_is_blue_verified': '1',
            'include_ext_verified_type': '1',
            'include_ext_profile_image_shape': '1',
            'skip_status': '1',
            'user_id': user_id,
        }
        headers = {
            'content-type': 'application/x-www-form-urlencoded',
        }
        response, response_json = await self.request("POST", url, params=params, headers=headers)
        return bool(response_json)

    async def follow(self, user_id: str | int) -> bool:
        return await self._follow_action("create", user_id)

    async def unfollow(self, user_id: str | int) -> bool:
        return await self._follow_action("destroy", user_id)

    async def _interact_with_tweet(self, action: str, tweet_id: int) -> dict:
        url, query_id = self._action_to_url(action)
        json_payload = {
            'variables': {
                'tweet_id': tweet_id,
                'dark_request': False
            },
            'queryId': query_id
        }
        response, response_json = await self.request("POST", url, json=json_payload)
        return response_json

    async def repost(self, tweet_id: int) -> int:
        """
        Repost (retweet)

        :return: Tweet ID
        """
        response_json = await self._interact_with_tweet('CreateRetweet', tweet_id)
        retweet_id = int(response_json['data']['create_retweet']['retweet_results']['result']['rest_id'])
        return retweet_id

    async def like(self, tweet_id: int) -> bool:
        response_json = await self._interact_with_tweet('FavoriteTweet', tweet_id)
        is_liked = response_json['data']['favorite_tweet'] == 'Done'
        return is_liked

    async def unlike(self, tweet_id: int) -> dict:
        response_json = await self._interact_with_tweet('UnfavoriteTweet', tweet_id)
        is_unliked = 'data' in response_json and response_json['data']['unfavorite_tweet'] == 'Done'
        return is_unliked

    async def delete_tweet(self, tweet_id: int | str) -> bool:
        url, query_id = self._action_to_url('DeleteTweet')
        json_payload = {
            'variables': {
                'tweet_id': tweet_id,
                'dark_request': False,
            },
            'queryId': query_id,
        }
        response, response_json = await self.request("POST", url, json=json_payload)
        is_deleted = "data" in response_json and "delete_tweet" in response_json["data"]
        return is_deleted

    async def pin_tweet(self, tweet_id: str | int) -> bool:
        url = 'https://api.twitter.com/1.1/account/pin_tweet.json'
        data = {
            'tweet_mode': 'extended',
            'id': str(tweet_id),
        }
        headers = {
            'content-type': 'application/x-www-form-urlencoded',
        }
        response, response_json = await self.request("POST", url, headers=headers, data=data)
        is_pinned = bool(response_json["pinned_tweets"])
        return is_pinned

    async def _tweet(
            self,
            text: str = None,
            *,
            media_id: int | str = None,
            tweet_id_to_reply: str | int = None,
            attachment_url: str = None,
    ) -> int:
        url, query_id = self._action_to_url('CreateTweet')
        payload = {
            'variables': {
                'tweet_text': text if text is not None else "",
                'dark_request': False,
                'media': {
                    'media_entities': [],
                    'possibly_sensitive': False},
                'semantic_annotation_ids': [],
            },
            'features': {
                'tweetypie_unmention_optimization_enabled': True,
                'responsive_web_edit_tweet_api_enabled': True,
                'graphql_is_translatable_rweb_tweet_is_translatable_enabled': True,
                'view_counts_everywhere_api_enabled': True,
                'longform_notetweets_consumption_enabled': True,
                'tweet_awards_web_tipping_enabled': False,
                'longform_notetweets_rich_text_read_enabled': True,
                'longform_notetweets_inline_media_enabled': True,
                'responsive_web_graphql_exclude_directive_enabled': True,
                'verified_phone_label_enabled': False,
                'freedom_of_speech_not_reach_fetch_enabled': True,
                'standardized_nudges_misinfo': True,
                'tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled': False,
                'responsive_web_graphql_skip_user_profile_image_extensions_enabled': False,
                'responsive_web_graphql_timeline_navigation_enabled': True,
                'responsive_web_enhance_cards_enabled': False,
                'responsive_web_twitter_article_tweet_consumption_enabled': False,
                'responsive_web_media_download_video_enabled': False
            },
            'queryId': query_id,
        }
        if attachment_url:
            payload['variables']['attachment_url'] = attachment_url
        if tweet_id_to_reply:
            payload['variables']['reply'] = {
                'in_reply_to_tweet_id': str(tweet_id_to_reply),
                'exclude_reply_user_ids': [],
            }
        if media_id:
            payload['variables']['media']['media_entities'].append({'media_id': str(media_id), 'tagged_users': []})

        response, response_json = await self.request("POST", url, json=payload)
        tweet_id = response_json['data']['create_tweet']['tweet_results']['result']['rest_id']
        return tweet_id

    async def tweet(self, text: str, *, media_id: int | str = None) -> int:
        """
        :return: Tweet ID
        """
        return await self._tweet(text, media_id=media_id)

    async def reply(self, tweet_id: str | int, text: str, *, media_id: int | str = None) -> int:
        """
        :return: Tweet ID
        """
        return await self._tweet(text, media_id=media_id, tweet_id_to_reply=tweet_id)

    async def quote(self, tweet_url: str, text: str, *, media_id: int | str = None) -> int:
        """
        :return: Tweet ID
        """
        return await self._tweet(text, media_id=media_id, attachment_url=tweet_url)

    async def vote(self, tweet_id: int | str, card_id: int | str, choice_number: int) -> dict:
        """
        :return: Raw vote information
        """
        url = "https://caps.twitter.com/v2/capi/passthrough/1"
        params = {
            "twitter:string:card_uri": f"card://{card_id}",
            "twitter:long:original_tweet_id": str(tweet_id),
            "twitter:string:response_card_name": "poll2choice_text_only",
            "twitter:string:cards_platform": "Web-12",
            "twitter:string:selected_choice": str(choice_number),
        }
        response, response_json = await self.request("POST", url, params=params)
        return response_json

    async def _request_users(self, action: str, user_id: int | str, count: int) -> list[TwitterUserData]:
        url, query_id = self._action_to_url(action)
        variables = {
            'userId': str(user_id),
            'count': count,
            'includePromotedContent': False,
        }
        features = {
            "rweb_lists_timeline_redesign_enabled": True,
            "responsive_web_graphql_exclude_directive_enabled": True,
            "verified_phone_label_enabled": False,
            "creator_subscriptions_tweet_preview_api_enabled": True,
            "responsive_web_graphql_timeline_navigation_enabled": True,
            "responsive_web_graphql_skip_user_profile_image_extensions_enabled": False,
            "tweetypie_unmention_optimization_enabled": True,
            "responsive_web_edit_tweet_api_enabled": True,
            "graphql_is_translatable_rweb_tweet_is_translatable_enabled": True,
            "view_counts_everywhere_api_enabled": True,
            "longform_notetweets_consumption_enabled": True,
            "responsive_web_twitter_article_tweet_consumption_enabled": False,
            "tweet_awards_web_tipping_enabled": False,
            "freedom_of_speech_not_reach_fetch_enabled": True,
            "standardized_nudges_misinfo": True,
            "tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled": True,
            "longform_notetweets_rich_text_read_enabled": True,
            "longform_notetweets_inline_media_enabled": True,
            "responsive_web_media_download_video_enabled": False,
            "responsive_web_enhance_cards_enabled": False
        }
        params = {
            'variables': to_json(variables),
            'features': to_json(features),
        }
        response, response_json = await self.request("GET", url, params=params)

        users = []
        if 'result' in response_json['data']['user']:
            entries = response_json['data']['user']['result']['timeline']['timeline']['instructions'][-1]['entries']
            for entry in entries:
                if entry['entryId'].startswith('user'):
                    user_data_dict = entry["content"]["itemContent"]["user_results"]["result"]
                    users.append(TwitterUserData.from_raw_user_data(user_data_dict))
        return users

    async def request_followers(self, user_id: int | str = None, count: int = 10) -> list[TwitterUserData]:
        """
        :param user_id: Текущий пользователь, если не передан ID иного пользователя.
        :param count: Количество подписчиков.
        """
        if user_id:
            return await self._request_users('Followers', user_id, count)
        else:
            if not self.account.id:
                await self.request_user_data()
            return await self._request_users('Followers', self.account.id, count)

    async def request_followings(self, user_id: int | str = None, count: int = 10) -> list[TwitterUserData]:
        """
        :param user_id: Текущий пользователь, если не передан ID иного пользователя.
        :param count: Количество подписчиков.
        """
        if user_id:
            return await self._request_users('Following', user_id, count)
        else:
            if not self.account.id:
                await self.request_user_data()
            return await self._request_users('Following', self.account.id, count)

    async def _request_tweet_data(self, tweet_id: int) -> dict:
        action = 'TweetDetail'
        url, query_id = self._action_to_url(action)
        variables = {
            "focalTweetId": str(tweet_id),
            "with_rux_injections": False,
            "includePromotedContent": True,
            "withCommunity": True,
            "withQuickPromoteEligibilityTweetFields": True,
            "withBirdwatchNotes": True,
            "withVoice": True,
            "withV2Timeline": True,
        }
        features = {
            "rweb_lists_timeline_redesign_enabled": True,
            "responsive_web_graphql_exclude_directive_enabled": True,
            "verified_phone_label_enabled": False,
            "creator_subscriptions_tweet_preview_api_enabled": True,
            "responsive_web_graphql_timeline_navigation_enabled": True,
            "responsive_web_graphql_skip_user_profile_image_extensions_enabled": False,
            "tweetypie_unmention_optimization_enabled": True,
            "responsive_web_edit_tweet_api_enabled": True,
            "graphql_is_translatable_rweb_tweet_is_translatable_enabled": True,
            "view_counts_everywhere_api_enabled": True,
            "longform_notetweets_consumption_enabled": True,
            "tweet_awards_web_tipping_enabled": False,
            "freedom_of_speech_not_reach_fetch_enabled": True,
            "standardized_nudges_misinfo": True,
            "tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled": True,
            "longform_notetweets_rich_text_read_enabled": True,
            "longform_notetweets_inline_media_enabled": True,
            "responsive_web_enhance_cards_enabled": False,
        }
        params = {
            'variables': to_json(variables),
            'features': to_json(features),
        }
        response, response_json = await self.request("GET", url, params=params)
        return response_json

    async def _update_profile_image(self, type: Literal["banner", "image"], media_id: str | int) -> str:
        """
        :return: Image URL
        """
        url = f"https://api.twitter.com/1.1/account/update_profile_{type}.json"
        params = {
            'media_id': str(media_id),
            'include_profile_interstitial_type': '1',
            'include_blocking': '1',
            'include_blocked_by': '1',
            'include_followed_by': '1',
            'include_want_retweets': '1',
            'include_mute_edge': '1',
            'include_can_dm': '1',
            'include_can_media_tag': '1',
            'include_ext_has_nft_avatar': '1',
            'include_ext_is_blue_verified': '1',
            'include_ext_verified_type': '1',
            'include_ext_profile_image_shape': '1',
            'skip_status': '1',
            'return_user': 'true',
        }
        response, response_json = await self.request("POST", url, params=params)
        image_url = response_json[f"profile_{type}_url"]
        return image_url

    async def update_profile_avatar(self, media_id: int | str) -> str:
        """
        :return: Image URL
        """
        return await self._update_profile_image("image", media_id)

    async def update_profile_banner(self, media_id: int | str) -> str:
        """
        :return: Image URL
        """
        return await self._update_profile_image("banner", media_id)

    async def change_username(self, username: str) -> bool:
        url = "https://twitter.com/i/api/1.1/account/settings.json"
        data = {"screen_name": username}
        response, response_json = await self.request("POST", url, data=data)
        new_username = response_json["screen_name"]
        is_changed = new_username == username
        self.account.username = new_username
        return is_changed

    async def change_password(self, password: str) -> bool:
        """
        После изменения пароля обновляется auth_token!
        """
        if not self.account.password:
            raise ValueError(f"Specify the current password before changing it")

        url = "https://twitter.com/i/api/i/account/change_password.json"
        data = {
            "current_password": self.account.password,
            "password": password,
            "password_confirmation": password
        }
        response, response_json = await self.request("POST", url, data=data)
        is_changed = response_json["status"] == "ok"
        auth_token = response.cookies.get("auth_token", domain=".twitter.com")
        self.account.auth_token = auth_token
        self.account.password = password
        return is_changed

    async def update_profile(
            self,
            name: str = None,
            description: str = None,
            location: str = None,
            website: str = None,
    ) -> bool:
        """
        Locks an account!
        """
        if name is None and description is None:
            raise ValueError("Specify at least one param")

        url = "https://twitter.com/i/api/1.1/account/update_profile.json"
        headers = {"content-type": "application/x-www-form-urlencoded"}
        # Создаем словарь data, включая в него только те ключи, для которых значения не равны None
        data = {k: v for k, v in [
            ("name", name),
            ("description", description),
            ("location", location),
            ("url", website),
        ] if v is not None}
        response, response_json = await self.request("POST", url, headers=headers, data=data)
        # Проверяем, что все переданные параметры соответствуют полученным
        is_updated = all(response_json.get(key) == value for key, value in data.items() if key != "url")
        if website: is_updated &= URL(website) == URL(response_json["entities"]["url"]["urls"][0]["expanded_url"])
        await self.establish_status()  # Изменение данных профиля часто замораживает аккаунт
        return is_updated

    async def establish_status(self):
        url = "https://twitter.com/i/api/1.1/account/update_profile.json"
        try:
            await self.request("POST", url)
        except BadTwitterAccount:
            pass

    async def update_birthdate(
            self,
            day: int,
            month: int,
            year: int,
            visibility: Literal["self", "mutualfollow"] = "self",
            year_visibility: Literal["self"] = "self",
    ) -> bool:
        url = "https://twitter.com/i/api/1.1/account/update_profile.json"
        headers = {"content-type": "application/x-www-form-urlencoded"}
        data = {
            "birthdate_day": day,
            "birthdate_month": month,
            "birthdate_year": year,
            "birthdate_visibility": visibility,
            "birthdate_year_visibility": year_visibility,
        }
        response, response_json = await self.request("POST", url, headers=headers, data=data)
        birthdate_data = response_json["extended_profile"]["birthdate"]
        is_updated = all((
            birthdate_data["day"] == day,
            birthdate_data["month"] == month,
            birthdate_data["year"] == year,
            birthdate_data["visibility"] == visibility,
            birthdate_data["year_visibility"] == year_visibility,
        ))
        return is_updated

    async def send_message(self, user_id: int | str, text: str) -> dict:
        """
        :return: Event data
        """
        url = "https://api.twitter.com/1.1/direct_messages/events/new.json"
        payload = {"event": {
            "type": "message_create",
            "message_create": {
                "target": {
                    "recipient_id": user_id
                }, "message_data": {
                    "text": text}
            }
        }}
        response, response_json = await self.request("POST", url, json=payload)
        event_data = response_json["event"]
        return event_data

    async def request_messages(self) -> list[dict]:
        """
        :return: Messages data
        """
        url = 'https://twitter.com/i/api/1.1/dm/inbox_initial_state.json'
        params = {
            'nsfw_filtering_enabled': 'false',
            'filter_low_quality': 'false',
            'include_quality': 'all',
            'include_profile_interstitial_type': '1',
            'include_blocking': '1',
            'include_blocked_by': '1',
            'include_followed_by': '1',
            'include_want_retweets': '1',
            'include_mute_edge': '1',
            'include_can_dm': '1',
            'include_can_media_tag': '1',
            'include_ext_has_nft_avatar': '1',
            'include_ext_is_blue_verified': '1',
            'include_ext_verified_type': '1',
            'include_ext_profile_image_shape': '1',
            'skip_status': '1',
            'dm_secret_conversations_enabled': 'false',
            'krs_registration_enabled': 'true',
            'cards_platform': 'Web-12',
            'include_cards': '1',
            'include_ext_alt_text': 'true',
            'include_ext_limited_action_results': 'true',
            'include_quote_count': 'true',
            'include_reply_count': '1',
            'tweet_mode': 'extended',
            'include_ext_views': 'true',
            'dm_users': 'true',
            'include_groups': 'true',
            'include_inbox_timelines': 'true',
            'include_ext_media_color': 'true',
            'supports_reactions': 'true',
            'include_ext_edit_control': 'true',
            'include_ext_business_affiliations_label': 'true',
            'ext': 'mediaColor,altText,mediaStats,highlightedLabel,hasNftAvatar,voiceInfo,birdwatchPivot,superFollowMetadata,unmentionInfo,editControl',
        }
        response, response_json = await self.request("GET", url, params=params)
        messages = [entry["message"] for entry in response_json["inbox_initial_state"]["entries"] if "message" in entry]
        return messages



from typing import Any

from curl_cffi import requests


__all__ = [
    "DiscordException",
    "HTTPException",
    "BadRequest",
    "CaptchaRequired",
    "Unauthorized",
    "Forbidden",
    "NotFound",
    "RateLimited",
    "DiscordServerError",
]


class DiscordException(Exception):
    pass


def _flatten_error_dict(d: dict[str, Any], key: str = "") -> dict[str, str]:
    items: list[tuple[str, str]] = []
    for k, v in d.items():
        new_key = key + '.' + k if key else k

        if isinstance(v, dict):
            try:
                _errors: list[dict[str, Any]] = v["_errors"]
            except KeyError:
                items.extend(_flatten_error_dict(v, new_key).items())
            else:
                items.append((new_key, " ".join(x.get("message", "") for x in _errors)))
        else:
            items.append((new_key, v))

    return dict(items)


class HTTPException(DiscordException):
    """Exception that's raised when an HTTP request operation fails
    """

    def __init__(
            self,
            response: requests.Response,
            data: str | dict[str, Any] | None,
            custom_exception_message: str = None,
    ):
        self.response = response
        if isinstance(data, dict):
            self.code = data.get("code", 0)
            base = data.get("message", "")
            errors = data.get('errors')
            self._errors: dict[str, Any] = errors
            if errors:
                errors = _flatten_error_dict(errors)
                helpful = "\n".join("In %s: %s" % t for t in errors.items())
                self.text = base + '\n' + helpful
            else:
                self.text = base
        else:
            self.text = data or ""
            self.code = 0

        exception_message = f"{self.response.status_code} ({self.code})"
        if len(self.text):
            exception_message = f"{exception_message}: {self.text}"
        super().__init__(custom_exception_message or exception_message)


class BadRequest(HTTPException):
    """Exception raised for a 400 HTTP status code.
    """
    pass


class CaptchaRequired(BadRequest):
    """Exception raised for a 400 HTTP status code and captcha bypass required.
    """

    def __init__(self, response: requests.Response, data: dict[str, Any]):
        self.sitekey = data["captcha_sitekey"]
        self.rqdata = data["captcha_rqdata"]
        self.rqtoken = data["captcha_rqtoken"]
        self.service = data["captcha_service"]
        super().__init__(
            response, data,
            custom_exception_message=f"You need to solve {self.service} to perform this action."
                                     f"\nBinding a phone number can reduce the chance of captchas appearing.",
        )


class Unauthorized(HTTPException):
    """Exception raised for a 401 HTTP status code.
    """
    pass


class Forbidden(HTTPException):
    """Exception raised for a 403 HTTP status code.
    """
    pass


class NotFound(HTTPException):
    """Exception raised for a 404 HTTP status code.
    """
    pass


class RateLimited(HTTPException):
    """Exception raised for a 429 HTTP status code.

    Attributes
    ------------
    retry_after:
        The amount of seconds that the client should wait before retrying
        the request.
    """

    def __init__(
            self,
            response: requests.Response,
            data: str | dict[str, Any] | None,
    ):
        self.retry_after = data["retry_after"]
        super().__init__(
            response, data,
            custom_exception_message=f"Rate limited. Retry in {self.retry_after:.2f} seconds.",
        )


class DiscordServerError(HTTPException):
    """Exception raised for a 5xx HTTP status code.
    """
    pass



import datetime
from functools import wraps
from random import choice
from typing import Any, Literal
from time import time
import string

from curl_cffi import requests
from yarl import URL

from .x_properties import create_guild_x_context_properties, create_x_super_properties, encode_x_properties
from .account import DiscordAccount, DiscordAccountStatus
from .errors import (
    HTTPException,
    BadRequest,
    Unauthorized,
    CaptchaRequired,
    RateLimited,
    Forbidden,
    NotFound,
    DiscordServerError,
)
from ..base import BaseClient


def generate_nonce() -> str:
    return str((int(time()) * 1000 - 1420070400000) * 4194304)


def generate_session_id() -> str:
    return "".join(choice(string.ascii_letters + string.digits) for _ in range(32))


def json_or_text(response: requests.Response) -> dict[str, Any] or str:
    try:
        if response.headers["content-type"] == "application/json":
            return response.json()
    except KeyError:
        # Thanks Cloudflare
        pass

    return response.text


class DiscordClient(BaseClient):
    BASE_URL = f"https://discord.com"
    API_VERSION = 9
    BASE_API_URL = f"{BASE_URL}/api/v{API_VERSION}"
    DEFAULT_HEADERS = {
        "authority": "discord.com",
        "origin": "https://discord.com",
        "x-debug-options": "bugReporterEnabled",
    }
    CLIENT_BUILD_NUMBER = 255423

    def __init__(
            self,
            account: DiscordAccount,
            *,
            locale: str = "en-US",
            **session_kwargs,
    ):
        super().__init__(**session_kwargs)
        self.account = account
        self.x_super_properties = create_x_super_properties(self.session.user_agent, self.CLIENT_BUILD_NUMBER)
        self.session.cookies.set("locale", locale, "discord.com", "/")
        self.session.headers.update({
            "x-discord-locale": locale,
            "accept-language": f"{locale},{locale.split('-')[0]};q=0.9",
        })

    async def request(
            self,
            method,
            url,
            **kwargs,
    ) -> tuple[requests.Response, dict[str, Any] or str]:
        headers = kwargs["headers"] = kwargs.get("headers") or {}
        headers["authorization"] = self.account.auth_token
        headers["x-super-properties"] = self.x_super_properties

        response = await self.session.request(method, url, **kwargs)
        data = json_or_text(response)

        if response.status_code == 401:
            self.account.status = DiscordAccountStatus.BAD_TOKEN
            raise Unauthorized(response, data)

        if response.status_code == 404:
            raise NotFound(response, data)

        if response.status_code >= 500:
            raise DiscordServerError(response, data)

        self.account.status = DiscordAccountStatus.GOOD

        if response.status_code == 400:
            if data.get("captcha_key"):
                raise CaptchaRequired(response, data)
            raise BadRequest(response, data)

        if response.status_code == 403:
            # May be BAD_TOKEN
            raise Forbidden(response, data)

        if response.status_code == 429:
            if not response.headers.get("via") or isinstance(data, str):
                # Banned by Cloudflare more than likely.
                raise HTTPException(response, data)
            raise RateLimited(response, data)

        if not 200 <= response.status_code < 300:
            raise HTTPException(response, data)

        if "flags" in data and "public_flags" in data:
            flags_data = data["flags"] - data["public_flags"]

            if flags_data == 17592186044416:
                self.account.is_quarantined = True
            elif flags_data == 1048576:
                self.account.is_spammer = True
            elif flags_data == 17592186044416 + 1048576:
                self.account.is_spammer = True
                self.account.is_quarantined = True

        return response, data

    # APP - BINDING

    async def bind_app(
            self,
            *,
            client_id: str,
            scope: str,
            state: str = None,
            response_type: str = "code",
    ):
        """
        Binds an application to the user's account using OAuth2.
        :param client_id: Client ID of the application.
        :param scope: The scope of the access request.
        :param state: An optional state parameter.
        :param response_type: The type of response required, defaults to 'code'.
        :return: The authorization code as part of the binding process.
        """
        url = f"{self.BASE_API_URL}/oauth2/authorize"
        params = {
            "client_id": client_id,
            "response_type": response_type,
            "scope": scope,
        }
        if state:
            params["state"] = state
        payload = {
            "permissions": "0",
            "authorize": True,
        }
        response, data = await self.request("POST", url, json=payload, params=params)
        bind_url = URL(data["location"])
        bind_code = bind_url.query.get("code")
        return bind_code

    # GUILD

    async def send_guild_chat_message(
            self,
            guild_id: int | str,
            channel_id: int | str,
            message: str,
    ) -> int:
        """
        Sends a message to a specified guild channel.
        :param guild_id: ID of the guild.
        :param channel_id: ID of the channel in the guild.
        :param message: The message content to send.
        :return: The ID of the sent message.
        """
        url = f"{self.BASE_API_URL}/channels/{channel_id}/messages"
        headers = {
            "referer": f"{self.BASE_URL}/channels/{guild_id}/{channel_id}",
        }
        payload = {
            "mobile_network_type": "unknown",
            "content": message,
            "nonce": generate_nonce(),
            "tts": False,
            "flags": 0,
        }
        response, data = await self.request("POST", url, json=payload, headers=headers)
        message_id = data["id"]
        return message_id

    async def send_reaction(
            self,
            channel_id: int | str,
            message_id: int | str,
            emoji: str,
    ):
        """
        Sends a reaction to a specific message in a channel.
        :param channel_id: ID of the channel.
        :param message_id: ID of the message to react to.
        :param emoji: The emoji to use for the reaction.
        """
        url = f"{self.BASE_API_URL}/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me"
        await self.request("PUT", url)

    async def press_button(
            self,
            guild_id: int | str,
            channel_id: int | str,
            message_id: int | str,
            application_id: int | str,
            button_data: dict,
    ):
        """
        Simulates pressing a button in a message (such as for interactions).
        :param guild_id: ID of the guild.
        :param channel_id: ID of the channel.
        :param message_id: ID of the message containing the button.
        :param application_id: ID of the application associated with the button.
        :param button_data: Data of the button being pressed.
        :return: Response data from the button press.
        """
        url = f"{self.BASE_API_URL}/interactions"
        headers = {
            'authority': 'discord.com',
            'accept': '*/*',
            'content-type': 'application/json',
            'origin': 'https://discord.com',
            'referer': f'https://discord.com/channels/{guild_id}/{channel_id}',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Windows"',
            'sec-fetch-dest': 'empty',
            'sec-fetch-mode': 'cors',
            'sec-fetch-site': 'same-origin',
            'x-debug-options': 'bugReporterEnabled',
            'x-discord-locale': 'en-US',
        }
        payload = {
            'type': 3,
            'nonce': generate_nonce(),
            'guild_id': guild_id,
            'channel_id': channel_id,
            'message_flags': 0,
            'message_id': message_id,
            'application_id': application_id,
            'session_id': generate_session_id(),
            'data': {
                'component_type': button_data['type'],
                'custom_id': button_data['custom_id'],
            }
        }
        response, data = await self.request("POST", url, headers=headers, json=payload)
        return data

    async def request_messages(
            self,
            channel_id: int | str,
            limit: int = 50,
            before_date: str = None,
            around_message_id: int | str = None,
            after_message_id: int | str = None,
    ) -> list[dict]:
        """
        Retrieves a list of messages from a specified channel.
        :param channel_id: ID of the channel.
        :param limit: The number of messages to retrieve (between 1 and 100).
        :param before_date: Get messages before this date (snowflake format).
        :param around_message_id: Get messages around this message ID.
        :param after_message_id: Get messages after this message ID.
        :return: A list of message data.
        """
        url = f"{self.BASE_API_URL}/channels/{channel_id}/messages"
        params = {"limit": limit}
        if before_date: params["before"] = before_date
        if around_message_id: params["around"] = around_message_id
        if after_message_id: params["after"] = after_message_id
        response, data = await self.request("GET", url, params=params)
        return data

    async def request_message(self, channel_id: int | str, message_id: int | str) -> dict:
        """
        Retrieves a specific message from a channel.
        :param channel_id: ID of the channel.
        :param message_id: ID of the message to retrieve.
        :return: Data of the requested message.
        """
        messages = await self.request_messages(channel_id, limit=1, around_message_id=message_id)
        return messages[0]

    async def request_invite_data(self, invite_code: str) -> dict:
        """
        Retrieves data for a specific invite code.
        :param invite_code: The invite code to retrieve data for.
        :return: Data associated with the invite code.
        """
        url = f"{self.BASE_API_URL}/invites/{invite_code}"
        params = {
            "with_counts": "true",
            "with_expiration": "true",
        }
        headers = {'referer': f'https://discord.com/invite/{invite_code}'}

        response, data = await self.request("GET", url, headers=headers, params=params)
        return data

    @staticmethod
    def check_cf_cookies(func):
        @wraps(func)
        async def wrapper(self, *args, **kwargs):
            # if not all(key in self.session.cookies for key in ("__cfruid", "__dcfduid")):
            if {"__cfruid", "__dcfduid"} <= self.session.cookies.keys():
                return await func(self, *args, **kwargs)
            else:
                raise ValueError("This action needs Cloudflare Cookies (__cfruid, __dcfduid) to bypass the captcha.")
        return wrapper

    @check_cf_cookies
    async def join_guild(
            self,
            invite_code: str,
            guild_id: int | str,
            channel_id: int | str,
            channel_type: int = 0,
            captcha_response: str = None,
            captcha_rqtoken: str = None,
    ) -> dict:
        """
        Joins a guild using an invite code.
        :param invite_code: The invite code of the guild to join.
        :param guild_id: The ID of the guild.
        :param channel_id: The ID of the channel in the guild.
        :param channel_type: The type of the channel.
        :param captcha_response: CAPTCHA response if required.
        :param captcha_rqtoken: CAPTCHA request token if required.
        :return: Response data after joining the guild.
        """
        url = f"{self.BASE_API_URL}/invites/{invite_code}"
        payload = {"session_id": None}
        headers = {
            'referer': f'https://discord.com/invite/{invite_code}',
            'x-context-properties': create_guild_x_context_properties(guild_id, channel_id, channel_type),
        }
        if captcha_response and captcha_rqtoken:
            headers["x-captcha-key"] = captcha_response
            headers["x-captcha-rqtoken"] = captcha_rqtoken

        response, data = await self.request("POST", url, headers=headers, json=payload)
        return data

    async def join_guild_with_invite_data(
            self,
            invite_data: dict,
            captcha_response: str = None,
            captcha_rqtoken: str = None,
    ) -> dict:
        """
        Joins a guild using invite data.
        :param invite_data: Dictionary containing invite data, including code, guild_id, and channel.
        :param captcha_response: CAPTCHA response if required.
        :param captcha_rqtoken: CAPTCHA request token if required.
        :return: Response data after joining the guild.
        """
        return await self.join_guild(
            invite_data["code"],
            invite_data["guild_id"],
            invite_data["channel"]["id"],
            channel_type=invite_data["channel"]["type"],
            captcha_response=captcha_response,
            captcha_rqtoken=captcha_rqtoken,
        )

    async def leave_guild(self, guild_id: int | str):
        """
        Leaves a guild.
        :param guild_id: The ID of the guild to leave.
        """
        url = f"{self.BASE_API_URL}/users/@me/guilds/{guild_id}"
        payload = {"lurking": False}
        await self.request("DELETE", url, json=payload)

    async def _request_guild_rules_form(
            self,
            invite_code: str,
            guild_id: int | str,
    ):
        url = f"{self.BASE_API_URL}/guilds/{guild_id}/member-verification"
        params = {
            "with_guild": "false",
            "invite_code": invite_code,
        }
        response, data = await self.request("GET", url, params=params)
        return data

    async def _agree_guild_rules(
            self,
            guild_id: int | str,
            channel_id: int | str,
            rules_form: dict,
    ):
        url = f"{self.BASE_API_URL}/guilds/{guild_id}/requests/@me"
        headers = {
            "referrer": f'https://discord.com/channels/{guild_id}/{channel_id}'
        }
        form_fields = rules_form["form_fields"][0].copy()
        form_fields["response"] = True
        payload = {
            "version": rules_form["version"],
            "form_fields": [form_fields],
        }
        response, data = await self.request("PUT", url, headers=headers, json=payload)
        return data

    async def agree_guild_rules(
            self,
            invite_code: str,
            guild_id: int | str,
            channel_id: int | str,
    ):
        """
        Agrees to the rules of a guild based on an invite code.
        :param invite_code: The invite code of the guild.
        :param guild_id: The ID of the guild.
        :param channel_id: The ID of the channel in the guild.
        :return: Response data after agreeing to the guild rules.
        """
        rules_form = await self._request_guild_rules_form(invite_code, guild_id)
        return await self._agree_guild_rules(guild_id, channel_id, rules_form)

    async def agree_guild_rules_with_invite_data(self, invite_data: dict):
        """
        Agrees to the rules of a guild using invite data.
        :param invite_data: A dictionary containing invite data, including code, guild_id, and channel.
        :return: Response data after agreeing to the guild rules.
        """
        return await self.agree_guild_rules(
            invite_data["code"],
            invite_data["guild_id"],
            invite_data["channel"]["id"]
        )

    async def create_invite_code(
            self,
            channel_id: int | str,
            *,
            max_age_seconds: int = 0,
            max_users: int = 0,
            grant_temporary_membership: bool = False,
            invite_code_to_validate: str = None,
            target_type: str = None,
    ) -> dict:
        """
        Creates an invite code for a specific channel.
        :param channel_id: ID of the channel to create the invite for.
        :param max_age_seconds: Duration in seconds for which the invite is valid.
        :param max_users: Maximum number of users that can use the invite.
        :param grant_temporary_membership: If true, users who join via the invite are granted temporary membership.
        :param invite_code_to_validate: An existing invite code to validate against.
        :param target_type: The type of target for the invite.
        :return: Data about the created invite code.
        """
        url = f"{self.BASE_API_URL}/channels/{channel_id}/invites"
        payload = {
            "max_age": max_age_seconds,
            "max_users": max_users,
            "temporary": grant_temporary_membership,
            "target_type": target_type
        }
        headers = {
            "x-context-properties": encode_x_properties({"location": "Hub Sidebar"}),
        }
        if invite_code_to_validate: payload["validate"] = invite_code_to_validate
        response, data = await self.request("POST", url, headers=headers, json=payload)
        return data

    async def delete_invite_code(self, invite_code: str):
        """
        Deletes an invite code. Note: This operation might not work and return a 404 error.
        :param invite_code: The invite code to be deleted.
        """
        url = f"{self.BASE_API_URL}/invites/{invite_code}"
        await self.request("DELETE", url)

    async def request_guild_invites(self, guild_id: int | str) -> int:
        """
        Retrieves a list of invite codes for a specific guild.
        :param guild_id: ID of the guild.
        :return: List of invite codes.
        """
        url = f"{self.BASE_API_URL}/guilds/{guild_id}/invites"
        response, data = await self.request("GET", url)
        return data

    async def reqeust_channel_invites(self, channel_id: int | str) -> int:
        """
        Retrieves a list of invite codes for a specific channel.
        :param channel_id: ID of the channel.
        :return: List of invite codes.
        """
        url = f"{self.BASE_API_URL}/channels/{channel_id}/invites"
        response, data = await self.request("GET", url)
        return data

    async def x_track_get_request(self, url, **request_kwargs):
        """
        Sends a GET request with 'x-track' header. This might be necessary for some endpoints.
        :param url: URL for the GET request.
        :param request_kwargs: Additional kwargs for the request.
        :return: Response data.
        """
        request_kwargs["headers"] = request_kwargs.get("headers", {})
        request_kwargs["headers"]["x-track"] = self.x_super_properties
        return await self.request("GET", url, **request_kwargs)

    async def request_guilds(self, with_counts: bool = True):
        """
        Retrieves a list of guilds the user is a member of.
        :param with_counts: Boolean to include count information.
        :return: List of guilds.
        """
        url = f"{self.BASE_API_URL}/users/@me/guilds"
        params = {"with_count": "true"} if with_counts else None
        response, data = await self.x_track_get_request(url, params=params)
        return data

    async def request_guild_channels(self, guild_id: int | str):
        """
        Retrieves a list of channels in a specific guild.
        :param guild_id: ID of the guild.
        :return: List of channels.
        """
        url = f"{self.BASE_API_URL}/guilds/{guild_id}/channels"
        response, data = await self.x_track_get_request(url)
        return data

    async def request_guild_roles(self, guild_id: int | str):
        """
        Retrieves a list of roles in a specific guild.
        :param guild_id: ID of the guild.
        :return: List of roles.
        """
        url = f"{self.BASE_API_URL}/guilds/{guild_id}/roles"
        response, data = await self.x_track_get_request(url)
        return data

    async def request_discoverable_guilds(self, offset: int | str = 0, limit: int | str = 48):
        """
        Retrieves a list of discoverable guilds.
        :param offset: Pagination offset.
        :param limit: Number of guilds to retrieve.
        :return: List of discoverable guilds.
        """
        url = f"{self.BASE_API_URL}/discoverable-guilds"
        params = {"offset": str(offset), "limit": str(limit)}
        response, data = await self.request("GET", url, params=params)
        return data

    async def request_guild_regions(self, guild_id):
        """
        Retrieves a list of available regions for a specific guild.
        :param guild_id: ID of the guild.
        :return: List of regions.
        """
        url = f"{self.BASE_API_URL}/guilds/{guild_id}/regions"
        response, data = await self.request("GET", url)
        return data

    # USER

    async def get_relationships(self):
        """
        Retrieves the current user's relationships.
        :return: List of relationships.
        """
        url = f"{self.BASE_API_URL}/users/@me/relationships"
        response, data = await self.request("GET", url)
        return data

    async def get_mutual_friends(self, user_id: str):
        """
        Retrieves mutual friends with a specified user.
        :param user_id: ID of the user to check mutual friends with.
        :return: List of mutual friends.
        """
        url = f"{self.BASE_API_URL}/users/{user_id}/relationships"
        response, data = await self.request("GET", url)
        return data

    async def request_friend(self, user: str):
        """
        Sends a friend request to a user or accepts a pending one.
        :param user: Username with discriminator (e.g., 'User#1234') or user ID.
        :return: Response data.
        """
        if "#" in user:
            username, discriminator = user.split("#")
            url = f"{self.BASE_API_URL}/users/@me/relationships"
            payload = {"username": username, "discriminator": int(discriminator)}
            response, data = await self.request("POST", url, json=payload)
        else:
            url = f"{self.BASE_API_URL}/users/@me/relationships/{user}"
            response, data = await self.request("PUT", url)
        return data

    async def accept_friend(self, user_id: str, location: str):
        """
        Accepts a friend request.
        :param user_id: ID of the user whose friend request is being accepted.
        :param location: Context location for the request.
        :return: Response data.
        """
        url = f"{self.BASE_API_URL}/users/@me/relationships/{user_id}"
        headers = {"X-Context-Properties": create_guild_x_context_properties(location, user_id)}
        response, data = await self.request("PUT", url, headers=headers)
        return data

    async def remove_relationship(self, user_id: str, location: str):
        """
        Removes a relationship with a user.
        :param user_id: ID of the user whose relationship is being removed.
        :param location: Context location for the request.
        :return: Response data.
        """
        url = f"{self.BASE_API_URL}/users/@me/relationships/{user_id}"
        headers = {"X-Context-Properties": create_guild_x_context_properties(location, user_id)}
        response, data = await self.request("DELETE", url, headers=headers)
        return data

    async def block_user(self, user_id: str, location: str):
        """
        Blocks a user.
        :param user_id: ID of the user to block.
        :param location: Context location for the request.
        :return: Response data.
        """
        url = f"{self.BASE_API_URL}/users/@me/relationships/{user_id}"
        body = {"type": 2}
        headers = {"X-Context-Properties": create_guild_x_context_properties(location, user_id)}
        response, data = await self.request("PUT", url, json=body, headers=headers)
        return data

    async def get_profile(self, user_id: str, with_mutual_guilds: bool = None, guild_id: int | str = None):
        """
        Retrieves a user's profile.
        :param user_id: ID of the user whose profile is being retrieved.
        :param with_mutual_guilds: Include mutual guilds in the response.
        :param guild_id: Specify a guild ID to check for mutual membership.
        :return: User's profile data.
        """
        url = f"{self.BASE_API_URL}/users/{user_id}/profile"
        params = {}
        if with_mutual_guilds is not None:
            params["with_mutual_guilds"] = str(with_mutual_guilds).lower()
        if guild_id is not None:
            params["guild_id"] = str(guild_id)
        response, data = await self.request("GET", url, params=params)
        return data

    async def request_user_data(self, with_analytics_token: bool = False):
        """
        Retrieves the current user's information.
        :param with_analytics_token: Include analytics token in the response.
        :return: Current user's data.
        """
        url = f"{self.BASE_API_URL}/users/@me"
        params = {}
        if with_analytics_token is not None:
            params["with_analytics_token"] = str(with_analytics_token).lower()
        response, data = await self.request("GET", url, params=params)
        self.account.username = data["username"]
        self.account.name = data["global_name"]
        self.account.email = data["email"]
        self.account.phone = data["phone"]
        return data

    async def get_user_affinities(self):
        """
        Retrieves the current user's user affinities.
        :return: List of user affinities.
        """
        url = f"{self.BASE_API_URL}/users/@me/affinities/users"
        response, data = await self.request("GET", url)
        return data

    async def get_guild_affinities(self):
        """
        Retrieves the current user's guild affinities.
        :return: List of guild affinities.
        """
        url = f"{self.BASE_API_URL}/users/@me/affinities/guilds"
        response, data = await self.request("GET", url)
        return data

    async def get_mentions(self, limit: int, role_mentions: bool, everyone_mentions: bool):
        """
        Retrieves mentions for the current user.
        :param limit: Number of mentions to retrieve.
        :param role_mentions: Include role mentions.
        :param everyone_mentions: Include @everyone mentions.
        :return: List of mentions.
        """
        url = f"{self.BASE_API_URL}/users/@me/mentions"
        params = {
            "limit": limit,
            "roles": str(role_mentions).lower(),
            "everyone": str(everyone_mentions).lower()
        }
        response, data = await self.request("GET", url, params=params)
        return data

    async def remove_mention_from_inbox(self, message_id: str):
        """
        Removes a mention from the user's inbox.
        :param message_id: ID of the message to remove.
        :return: Response data.
        """
        url = f"{self.BASE_API_URL}/users/@me/mentions/{message_id}"
        response, data = await self.request("DELETE", url)
        return data

    # USER - SETTINGS

    async def _set_user_data(
            self,
            payload: dict,
            captcha_response: str = None,
            captcha_rqtoken: str = None,
    ) -> dict:
        """
        :return: User data.
        """
        url = f"{self.BASE_API_URL}/users/@me"
        headers = None
        if captcha_response and captcha_rqtoken:
            headers = {
                "x-captcha-key": captcha_response,
                "x-captcha-rqtoken": captcha_rqtoken,
            }
        response, data = await self.request("PATCH", url, headers=headers, json=payload)
        return data

    @check_cf_cookies
    async def change_username(
            self,
            username: str,
            captcha_response: str = None,
            captcha_rqtoken: str = None,
    ) -> dict:
        """
        Changes the username of the current user account.
        :param username: New username to be set.
        :param captcha_response: CAPTCHA response if required.
        :param captcha_rqtoken: CAPTCHA request token if required.
        :return: Updated user data after changing the username.
        :raises ValueError: If the current password is not specified in the account.
        """
        if not self.account.password:
            raise ValueError("Specify the current password before changing username.")

        payload = {
            "username": username,
            "password": self.account.password,
        }
        data = await self._set_user_data(payload, captcha_response, captcha_rqtoken)
        self.account.username = username
        return data

    async def set_name(
            self,
            name: str,
            captcha_response: str = None,
            captcha_rqtoken: str = None,
    ) -> dict:
        """
        Sets the global name for the user account.
        :param name: New global name to be set.
        :param captcha_response: CAPTCHA response if required.
        :param captcha_rqtoken: CAPTCHA request token if required.
        :return: Updated user data after setting the global name.
        """
        payload = {"global_name": name}
        data = await self._set_user_data(payload, captcha_response, captcha_rqtoken)
        self.account.name = name
        return data

    async def set_password(
            self,
            new_password: str,
    ):
        """
        Changes the password of the current user account.
        :param new_password: New password to be set.
        :return: Updated user data after changing the password.
        :raises ValueError: If the current password is not specified in the account.
        """
        if not self.account.password:
            raise ValueError("Specify the current password before changing it.")

        url = f"{self.BASE_API_URL}/users/@me"
        headers = {
            'connection': 'keep-alive',
            'referer': url,
        }
        payload = {
            'password': self.account.password,
            'new_password': new_password,
        }
        response, data = await self.request("PATCH", url, headers=headers, json=payload)
        self.account.auth_token = data["token"]
        self.account.password = new_password
        return data

    async def set_avatar(self, encoded_image_base64: str, image_ext: Literal['png', 'gif', 'jpeg'] = 'jpeg'):
        """
        Sets the user's avatar.
        :param encoded_image_base64: Base64 encoded image string.
        :param image_ext: Image extension, defaults to 'jpeg'.
        :return: Response data.
        """
        body = {"avatar": f"data:image/{image_ext};base64,{encoded_image_base64}"}
        response, data = await self._set_user_data(body)
        return data

    async def set_profile_color(self, color: int):
        """
        Sets the user's profile color.
        :param color: The color value to set.
        :return: Response data.
        """
        body = {"accent_color": color}
        response, data = await self._set_user_data(body)
        return data

    async def set_email(self, email: str, password: str):
        """
        Sets the user's email.
        :param email: New email address.
        :param password: Current password for verification.
        :return: Response data.
        """
        body = {"email": email, "password": password}
        response, data = await self._set_user_data(body)
        return data

    async def set_discriminator(self, discriminator: int, password: str):
        """
        Sets the user's discriminator.
        :param discriminator: New discriminator value.
        :param password: Current password for verification.
        :return: Response data.
        """
        body = {"password": password, "discriminator": discriminator}
        response, data = await self._set_user_data(body)
        return data

    async def set_about_me(self, bio: str):
        """
        Sets the user's bio. Note: Requires beta program access.
        :param bio: Bio text to set.
        :return: Response data.
        """
        body = {"bio": bio}
        response, data = await self._set_user_data(body)
        return data

    async def set_banner(self, encoded_image_base64: str):
        """
        Sets the user's banner.
        :param encoded_image_base64: Base64 encoded banner image.
        :return: Response data.
        """
        body = {"banner": f"data:image/png;base64,{encoded_image_base64}"}
        response, data = await self._set_user_data(body)
        return data

    async def enable_2fa(self, code: str, secret: str, password: str):
        """
        Enables 2FA for the user account.
        :param code: 2FA code.
        :param secret: 2FA secret.
        :param password: Current password for verification.
        :return: Response data including new token and backup codes.
        """
        url = f"{self.BASE_API_URL}/users/@me/mfa/totp/enable"
        body = {"code": code, "secret": secret, "password": password}
        response, data = await self.request("POST", url, json=body)
        return data

    async def disable_2fa(self, code: str):
        """
        Disables 2FA for the user account.
        :param code: 2FA code for verification.
        :return: Response data.
        """
        url = f"{self.BASE_API_URL}/users/@me/mfa/totp/disable"
        body = {"code": code}
        response, data = await self.request("POST", url, json=body)
        return data

    async def get_backup_codes(self, password: str, regenerate: bool):
        """
        Retrieves backup codes for 2FA.
        :param password: Current password for verification.
        :param regenerate: Whether to regenerate new backup codes.
        :return: Response data with backup codes.
        """
        url = f"{self.BASE_API_URL}/users/@me/mfa/codes"
        body = {"password": password, "regenerate": regenerate}
        response, data = await self.request("POST", url, json=body)
        return data

    async def disable_account(self, password: str):
        """
        Disables the user account.
        :param password: Current password for verification.
        :return: Response data.
        """
        url = f"{self.BASE_API_URL}/users/@me/disable"
        body = {"password": password}
        response, data = await self.request("POST", url, json=body)
        return data

    async def delete_account(self, password: str):
        """
        Deletes the user account.
        :param password: Current password for verification.
        :return: Response data.
        """
        url = f"{self.BASE_API_URL}/users/@me/delete"
        body = {"password": password}
        response, data = await self.request("POST", url, json=body)
        return data

    async def set_phone(self, number: str, reason: str):
        """
        Sets the user's phone number.
        :param number: Phone number to set.
        :param reason: Reason for the change.
        :return: Response data.
        """
        url = f"{self.BASE_API_URL}/users/@me/phone"
        body = {"phone": number, "change_phone_reason": reason}
        response, data = await self.request("POST", url, json=body)
        return data

    async def validate_phone(self, number: str, code: int, password: str):
        """
        Validates the phone number for the user account.
        :param number: Phone number to validate.
        :param code: Verification code sent to the phone.
        :param password: Current password for verification.
        :return: Response data.
        """
        url = f"{self.BASE_API_URL}/phone-verifications/verify"
        body = {"phone": number, "code": str(code)}
        response, data = await self.request("POST", url, json=body)

        url = f"{self.BASE_API_URL}/users/@me/phone"
        body = {"phone_token": data["token"], "password": password}
        response, data = await self.request("POST", url, json=body)
        return data

    # USER - SETTINGS - Privacy & Safety

    async def set_dm_scan_level(self, level: int):
        """
        Sets the explicit content filter level for direct messages.
        :param level: The level of the explicit content filter.
        :return: Response data after updating the settings.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"explicit_content_filter": level}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def allow_dms_from_server_members(self, allow: bool, disallowed_guild_ids: list):
        """
        Configures whether to allow direct messages from server members.
        :param allow: Boolean indicating if DMs are allowed.
        :param disallowed_guild_ids: List of guild IDs to restrict DMs from.
        :return: Response data after updating the settings.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"restricted_guilds": disallowed_guild_ids, "default_guilds_restricted": not allow}
        if not disallowed_guild_ids:
            body.pop("restricted_guilds")
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def allow_friend_requests_from(self, types: list):
        """
        Sets the preferences for receiving friend requests.
        :param types: List of types indicating who can send friend requests.
        :return: Response data after updating the settings.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"friend_source_flags": {"all": True, "mutual_friends": True, "mutual_guilds": True}}
        types = [i.lower().strip() for i in types]
        if "everyone" not in types:
            body["friend_source_flags"]["all"] = False
        if "mutual_friends" not in types:
            body["friend_source_flags"]["mutual_friends"] = False
        if "mutual_guilds" not in types:
            body["friend_source_flags"]["mutual_guilds"] = False
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def analytics_consent(self, grant: list, revoke: list):
        """
        Manages consent for analytics data.
        :param grant: List of analytics types to grant consent for.
        :param revoke: List of analytics types to revoke consent for.
        :return: Response data after updating the consent settings.
        """
        url = f"{self.BASE_API_URL}/users/@me/consent"
        body = {"grant": grant, "revoke": revoke}
        response, data = await self.request("POST", url, json=body)
        return data

    async def allow_screen_reader_tracking(self, allow: bool):
        """
        Configures tracking for screen readers.
        :param allow: Boolean indicating if screen reader tracking is allowed.
        :return: Response data after updating the settings.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"allow_accessibility_detection": allow}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def request_my_data(self):
        """
        Requests a copy of the user's data from Discord.
        :return: Response data indicating the status of the data request.
        """
        url = f"{self.BASE_API_URL}/users/@me/harvest"
        response, data = await self.request("POST", url)
        return data

    # USER - SETTINGS - Connections

    async def get_connected_accounts(self):
        """
        Retrieves the accounts connected to the user's Discord account.
        :return: List of connected accounts.
        """
        url = f"{self.BASE_API_URL}/users/@me/connections"
        response, data = await self.request("GET", url)
        return data

    async def get_connection_url(self, account_type: str):
        """
        Retrieves the authorization URL for a specific connection account type.
        :param account_type: The type of the account to get the connection URL for.
        :return: Authorization URL for the specified account type.
        """
        url = f"{self.BASE_API_URL}/connections/{account_type}/authorize"
        response, data = await self.request("GET", url)
        return data

    async def enable_connection_display_on_profile(self, account_type: str, account_username: str, enable: bool):
        """
        Enables or disables the display of a connected account on the user's profile.
        :param account_type: The type of the connected account.
        :param account_username: The username of the connected account.
        :param enable: Boolean to enable or disable the display on profile.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/connections/{account_type}/{account_username}"
        body = {"visibility": enable}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def enable_connection_display_on_status(self, account_type: str, account_username: str, enable: bool):
        """
        Enables or disables the display of a connected account on the user's status.
        :param account_type: The type of the connected account.
        :param account_username: The username of the connected account.
        :param enable: Boolean to enable or disable the display on status.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/connections/{account_type}/{account_username}"
        body = {"show_activity": enable}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def remove_connection(self, account_type: str, account_username: str):
        """
        Removes a connected account from the user's Discord account.
        :param account_type: The type of the connected account to remove.
        :param account_username: The username of the connected account to remove.
        :return: Response data after removing the connection.
        """
        url = f"{self.BASE_API_URL}/users/@me/connections/{account_type}/{account_username}"
        response, data = await self.request("DELETE", url)
        return data

    # USER - BILLING SETTINGS

    async def get_billing_history(self, limit: int):
        """
        Retrieves the billing history of the user.
        :param limit: The number of billing records to retrieve.
        :return: Billing history data.
        """
        url = f"{self.BASE_API_URL}/users/@me/billing/payments?limit={limit}"
        response, data = await self.request("GET", url)
        return data

    async def get_payment_sources(self):
        """
        Retrieves the payment sources associated with the user's account.
        :return: List of payment sources.
        """
        url = f"{self.BASE_API_URL}/users/@me/billing/payment-sources"
        response, data = await self.request("GET", url)
        return data

    async def get_billing_subscriptions(self):
        """
        Retrieves the billing subscriptions associated with the user's account.
        :return: List of billing subscriptions.
        """
        url = f"{self.BASE_API_URL}/users/@me/billing/subscriptions"
        response, data = await self.request("GET", url)
        return data

    async def get_stripe_client_secret(self):
        """
        Retrieves the Stripe client secret for adding new payment methods.
        Note: This method is for Stripe integration and requires Stripe API setup.
        :return: Stripe client secret data.
        """
        url = f"{self.BASE_API_URL}/users/@me/billing/stripe/setup-intents"
        response, data = await self.request("POST", url)
        return data

    # USER - APP SETTINGS - Appearance

    async def set_theme(self, theme: str):
        """
        Sets the theme of the user's Discord interface.
        :param theme: The theme to set ('light' or 'dark').
        :return: Response data after updating the theme setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"theme": theme.lower()}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def set_message_display(self, cozy_or_compact: str):
        """
        Sets the message display format in Discord.
        :param cozy_or_compact: 'Cozy' for cozy display format or 'Compact' for compact display format.
        :return: Response data after updating the message display setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        is_compact = cozy_or_compact.lower() == "compact"
        body = {"message_display_compact": is_compact}
        response, data = await self.request("PATCH", url, json=body)
        return data

    # USER - APP SETTINGS - Accessibility

    async def enable_gif_auto_play(self, enable: bool):
        """
        Enables or disables automatic GIF playback.
        :param enable: Boolean indicating whether to enable or disable GIF auto-play.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"gif_auto_play": enable}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def enable_animated_emoji(self, enable: bool):
        """
        Enables or disables animated emojis.
        :param enable: Boolean indicating whether to enable or disable animated emojis.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"animate_emoji": enable}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def set_sticker_animation(self, setting: str):
        """
        Sets the sticker animation preference.
        :param setting: String indicating the sticker animation setting ('always', 'interaction', 'never').
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        animation_settings = {"always": 0, "interaction": 1, "never": 2}
        body = {"animate_stickers": animation_settings.get(setting.lower(), 0)}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def enable_tts(self, enable: bool):
        """
        Enables or disables Text-To-Speech (TTS) commands.
        :param enable: Boolean indicating whether to enable or disable TTS commands.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"enable_tts_command": enable}
        response, data = await self.request("PATCH", url, json=body)
        return data

    # USER - APP SETTINGS - Text & Images

    async def enable_linked_image_display(self, enable: bool):
        """
        Enables or disables the display of images linked in chat.
        :param enable: Boolean indicating whether to enable or disable inline image display.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"inline_embed_media": enable}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def enable_image_display(self, enable: bool):
        """
        Enables or disables the display of images directly uploaded to chat.
        :param enable: Boolean indicating whether to enable or disable inline attachment media display.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"inline_attachment_media": enable}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def enable_link_preview(self, enable: bool):
        """
        Enables or disables link preview in chat messages.
        :param enable: Boolean indicating whether to enable or disable rendering of link embeds.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"render_embeds": enable}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def enable_reaction_rendering(self, enable: bool):
        """
        Enables or disables the rendering of message reactions.
        :param enable: Boolean indicating whether to enable or disable rendering of reactions.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"render_reactions": enable}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def enable_emoticon_conversion(self, enable: bool):
        """
        Enables or disables the conversion of emoticons to emojis.
        :param enable: Boolean indicating whether to enable or disable emoticon conversion.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"convert_emoticons": enable}
        response, data = await self.request("PATCH", url, json=body)
        return data

    # USER - APP SETTINGS - Notifications

    async def set_afk_timeout(self, timeout_seconds: int):
        """
        Sets the AFK (Away From Keyboard) timeout duration.
        :param timeout_seconds: The duration in seconds for AFK timeout.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"afk_timeout": timeout_seconds}
        response, data = await self.request("PATCH", url, json=body)
        return data

    # USER - APP SETTINGS - Language

    async def set_locale(self, locale: str):
        """
        Sets the user's language preference.
        :param locale: String representing the locale to set (e.g., 'en-US').
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"locale": locale}
        response, data = await self.request("PATCH", url, json=body)
        return data

    # USER - APP SETTINGS - Advanced

    async def enable_dev_mode(self, enable: bool):
        """
        Enables or disables Developer Mode in Discord settings.
        :param enable: Boolean indicating whether to enable or disable Developer Mode.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"developer_mode": enable}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def activate_application_test_mode(self, application_id: str):
        """
        Activates test mode for a specific application.
        :param application_id: The ID of the application to activate test mode for.
        :return: Response data after activating test mode.
        """
        url = f"{self.BASE_API_URL}/applications/{application_id}/skus"
        response, data = await self.request("GET", url)
        return data

    async def get_application_data(self, application_id: str, with_guild: bool):
        """
        Retrieves data for a specific application.
        :param application_id: The ID of the application.
        :param with_guild: Boolean indicating whether to include guild information.
        :return: Application data.
        """
        url = f"{self.BASE_API_URL}/applications/{application_id}/public?with_guild={str(with_guild).lower()}"
        response, data = await self.request("GET", url)
        return data

    # USER - ACTIVITY SETTINGS

    async def enable_activity_display(self, enable: bool, timeout: int = None):
        """
        Enables or disables the display of current activity status.
        :param enable: Boolean indicating whether to show the current game/activity.
        :param timeout: Optional timeout parameter for the request.
        :return: Response data after updating the activity display setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/settings"
        body = {"show_current_game": enable}
        response, data = await self.request("PATCH", url, json=body, timeout=timeout)
        return data

    # USER - OTHER SETTINGS - HypeSquad

    async def set_hypesquad(self, house: str):
        """
        Sets the HypeSquad house for the user.
        :param house: The HypeSquad house to join ('Bravery', 'Brilliance', 'Balance').
        :return: Response data after joining the HypeSquad house.
        """
        url = f"{self.BASE_API_URL}/hypesquad/online"
        house_ids = {"bravery": 1, "brilliance": 2, "balance": 3}
        body = {"house_id": house_ids.get(house.lower(), 1)}
        response, data = await self.request("POST", url, json=body)
        return data

    async def leave_hypesquad(self):
        """
        Leaves the HypeSquad.
        :return: Response data after leaving the HypeSquad.
        """
        url = f"{self.BASE_API_URL}/hypesquad/online"
        response, data = await self.request("DELETE", url)
        return data

    # USER - OTHER SETTINGS - Developer Options

    async def get_build_overrides(self):
        """
        Retrieves build override settings for the user.
        :return: Response data with build override settings.
        """
        url = "https://discord.com/__development/build_overrides"
        response, data = await self.request("GET", url)
        return data

    async def enable_source_maps(self, enable: bool):
        """
        Enables or disables source maps.
        :param enable: Boolean indicating whether to enable or disable source maps.
        :return: Response data after updating the setting.
        """
        url = "https://discord.com/__development/source_maps"
        method = "PUT" if enable else "DELETE"
        response, data = await self.request(method, url)
        return data

    # USER - OTHER SETTINGS - Notification Settings

    async def suppress_everyone_pings(self, guild_id: int | str, suppress: bool):
        """
        Suppresses or allows '@everyone' pings in a specific guild.
        :param guild_id: ID of the guild.
        :param suppress: Boolean indicating whether to suppress '@everyone' pings.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/guilds/{guild_id}/settings"
        body = {"suppress_everyone": suppress}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def suppress_role_mentions(self, guild_id: int | str, suppress: bool):
        """
        Suppresses or allows role mentions in a specific guild.
        :param guild_id: ID of the guild.
        :param suppress: Boolean indicating whether to suppress role mentions.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/guilds/{guild_id}/settings"
        body = {"suppress_roles": suppress}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def enable_mobile_push_notifications(self, guild_id: int | str, enable: bool):
        """
        Enables or disables mobile push notifications for a specific guild.
        :param guild_id: ID of the guild.
        :param enable: Boolean indicating whether to enable mobile push notifications.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/guilds/{guild_id}/settings"
        body = {"mobile_push": enable}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def set_channel_notification_overrides(self, guild_id: int | str, overrides: dict):
        """
        Sets channel-specific notification overrides for a guild.
        :param guild_id: ID of the guild.
        :param overrides: Dictionary of channel overrides.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/guilds/{guild_id}/settings"
        body = {"channel_overrides": overrides}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def set_message_notifications(self, guild_id: int | str, notifications: str):
        """
        Sets the message notification setting for a guild.
        :param guild_id: ID of the guild.
        :param notifications: Notification setting ('all messages', 'only mentions', 'nothing').
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/guilds/{guild_id}/settings"
        notification_types = {"all messages": 0, "only mentions": 1, "nothing": 2}
        body = {"message_notifications": notification_types.get(notifications.lower(), 1)}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def mute_guild(self, guild_id: int | str, mute: bool, duration: int = None):
        """
        Mutes or unmutes a guild for a specified duration.
        :param guild_id: ID of the guild.
        :param mute: Boolean indicating whether to mute the guild.
        :param duration: Duration in minutes to mute the guild, None for indefinite.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/guilds/{guild_id}/settings"
        body = {"muted": mute}
        if mute and duration is not None:
            end_time = datetime.datetime.now(datetime.UTC) + datetime.timedelta(minutes=duration)
            body["mute_config"] = {"selected_time_window": duration, "end_time": end_time.isoformat()}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def mute_dm(self, dm_id: int | str, mute: bool, duration: int = None):
        """
        Mutes or unmutes a direct message channel for a specified duration.
        :param dm_id: ID of the direct message channel.
        :param mute: Boolean indicating whether to mute the DM.
        :param duration: Duration in minutes to mute the DM, None for indefinite.
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/users/@me/guilds/%40me/settings"
        data = {"muted": mute}
        if mute and duration is not None:
            end_time = datetime.datetime.now(datetime.UTC) + datetime.timedelta(minutes=duration)
            data["mute_config"] = {"selected_time_window": duration, "end_time": end_time.isoformat()}
        body = {"channel_overrides": {str(dm_id): data}}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def set_thread_notifications(self, thread_id: int | str, notifications: str):
        """
        Sets the notification setting for a specific thread.
        :param thread_id: ID of the thread.
        :param notifications: Notification setting ('all messages', 'only mentions', 'nothing').
        :return: Response data after updating the setting.
        """
        url = f"{self.BASE_API_URL}/channels/{thread_id}/thread-members/@me/settings"
        notification_types = {"all messages": 0, "only mentions": 1, "nothing": 2}
        flags = 1 << (notification_types.get(notifications.lower(), 1) + 1)
        body = {"flags": flags}
        response, data = await self.request("PATCH", url, json=body)
        return data

    async def get_report_menu(self):
        """
        Retrieves the report menu for the first direct message.
        :return: Response data with report menu options.
        """
        url = f"{self.BASE_API_URL}/reporting/menu/first_dm"
        response, data = await self.request("GET", url)
        return data

    async def report_spam(self, channel_id: int | str, message_id: int | str, report_type: str, guild_id: int | str,
                          version: str, variant: str, language: str):
        """
        Reports a message as spam.
        :param channel_id: ID of the channel where the message is located.
        :param message_id: ID of the message to report.
        :param report_type: Type of report.
        :param guild_id: ID of the guild, if applicable.
        :param version: Version information for the report.
        :param variant: Variant of the report.
        :param language: Language setting for the report.
        :return: Response data after submitting the report.
        """
        url = f"{self.BASE_API_URL}/reporting/{report_type}"
        body = {
            "id": generate_nonce(),
            "version": version,
            "variant": variant,
            "language": language,
            "breadcrumbs": [7],
            "elements": {},
            "name": report_type,
            "channel_id": channel_id,
            "message_id": message_id,
        }
        if report_type in ('guild_directory_entry', 'stage_channel', 'guild'):
            body["guild_id"] = guild_id
        response, data = await self.request("POST", url, json=body)
        return data

    async def get_handoff_token(self, key: str):
        """
        Retrieves a handoff token.
        :param key: The key for which the handoff token is requested.
        :return: Response data with the handoff token.
        """
        url = f"{self.BASE_API_URL}/auth/handoff"
        body = {"key": key}
        response, data = await self.request("POST", url, json=body)
        return data

    async def invite_to_call(self, channel_id: int | str, user_ids: list):
        """
        Invites users to a call in a specific channel.
        :param channel_id: ID of the channel where the call is hosted.
        :param user_ids: List of user IDs to invite to the call.
        :return: Response data after sending the call invites.
        """
        url = f"{self.BASE_API_URL}/channels/{channel_id}/call/ring"
        body = {"recipients": user_ids}
        response, data = await self.request("POST", url, json=body)
        return data

    async def decline_call(self, channel_id: int | str):
        """
        Declines a call in a specific channel.
        :param channel_id: ID of the channel where the call is hosted.
        :return: Response data after declining the call.
        """
        url = f"{self.BASE_API_URL}/channels/{channel_id}/call/stop-ringing"
        response, data = await self.request("POST", url)
        return data

    # USER - Logout

    async def logout(self, provider: str, voip_provider: str):
        """
        Logs the user out of Discord.
        :param provider: The authentication provider.
        :param voip_provider: The VOIP provider.
        :return: Response data after logging out.
        """
        url = f"{self.BASE_API_URL}/auth/logout"
        body = {"provider": provider, "voip_provider": voip_provider}
        response, data = await self.request("POST", url, json=body)
        return data




from curl_cffi import requests


__all__ = [
    "GoogleAPIsException",
    "HTTPException",
    "BadRequest",
    "Unauthorized",
    "Forbidden",
    "NotFound",
    "GoogleAPIsServerError",
]


class GoogleAPIsException(Exception):
    pass


class HTTPException(GoogleAPIsException):
    """Exception that's raised when an HTTP request operation fails
    """

    def __init__(
            self,
            response: requests.Response,
            data: dict,
            custom_exception_message: str = None,
    ):
        self.response = response
        error_data = data["error"]
        self.message = error_data["message"]
        self.error_list = error_data["errors"]

        exception_message = f"{self.response.status_code} {self.message}"
        super().__init__(custom_exception_message or exception_message)


class BadRequest(HTTPException):
    """Exception raised for a 400 HTTP status code.
    """
    pass


class Unauthorized(HTTPException):
    """Exception raised for a 401 HTTP status code.
    """
    pass


class Forbidden(HTTPException):
    """Exception raised for a 403 HTTP status code.
    """
    pass


class NotFound(HTTPException):
    """Exception raised for a 404 HTTP status code.
    """
    pass


class GoogleAPIsServerError(HTTPException):
    """Exception raised for a 5xx HTTP status code.
    """
    pass



import random
import sys
import asyncio

from tqdm.asyncio import tqdm


def set_windows_selector_event_loop_policy():
    if sys.version_info >= (3, 8) and sys.platform.lower().startswith("win"):
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


def curry_async(async_func):
    async def curried(*args, **kwargs):
        def bound_async_func(*args2, **kwargs2):
            return async_func(*(args + args2), **{**kwargs, **kwargs2})

        return bound_async_func

    return curried


async def gather(*fs, max_tasks: int = None, delay: tuple[int, int] = (0, 0), **tqdm_kwargs):
    if max_tasks is None and delay == (0, 0):
        return await tqdm.gather(*fs, **tqdm_kwargs)

    semaphore = asyncio.Semaphore(max_tasks) if max_tasks is not None else None

    async def worker(fn):
        async def task():
            if delay != (0, 0):
                await asyncio.sleep(random.uniform(*delay))
            return await fn

        if semaphore:
            async with semaphore:
                return await task()
        else:
            return await task()

    return await tqdm.gather(*(worker(fn) for fn in fs), **tqdm_kwargs)



from bs4 import BeautifulSoup


def parse_oauth_html(html: str) -> dict:
    soup = BeautifulSoup(html, "lxml")
    result = {
        "oauth_token": soup.find("input", attrs={"name": "authenticity_token"}).get("value"),
        "authenticity_token": soup.find("input", attrs={"name": "authenticity_token"}).get("value"),
    }
    redirect_url_element = soup.find("a", text="click here to continue")
    if redirect_url_element: result["redirect_url"] = redirect_url_element.get("href")
    redirect_after_login_element = soup.find("input", attrs={"name": "redirect_after_login"})
    if redirect_after_login_element: result["redirect_after_login_url"] = redirect_after_login_element.get("value")
    return result


def remove_at_sign(username: str) -> str:
    if username.startswith("@"):
        return username[1:]
    return username


def tweet_url(username: str, tweet_id: int) -> str:
    """
    :return: Tweet URL
    """
    return f"https://x.com/{username}/status/{tweet_id}"



import asyncio
import time

from ..base import BaseClient
from .errors import SmsServiceError


class SmshubClient(BaseClient):
    BASE_API_URL = "https://smshub.org/stubs/handler_api.php"
    UNOFFICIAL_API_URL = "https://smshub.org/api.php"

    def __init__(self, key: str, **session_kwargs):
        super().__init__(**session_kwargs)
        self.key = key

    async def _request(self, method: str, url: str, **kwargs):
        params = kwargs["params"] = kwargs.get("params") or {}
        params["api_key"] = self.key
        response = await self.session.request(method, url, **kwargs)

        if response.text.startswith('ERROR'):
            raise SmsServiceError(response.text)

        return response

    async def request_balance(self) -> float:
        """https://smshub.org/ru/info#getBalance"""
        params = {'action': 'getBalance'}
        response = await self._request("GET", self.BASE_API_URL, params=params)

        _, balance_str = response.text.split(':')
        balance_float = float(balance_str)
        return balance_float

    async def request_numbers_status(self, country: str = None, operator: str = None) -> dict:
        params = {'action': 'getNumbersStatus'}
        if country:
            params['country'] = country
        if operator:
            params['operator'] = operator
        response = await self._request("GET", self.BASE_API_URL, params=params)
        return response.json()

    async def request_number(
            self,
            service: str,
            country: str = None,
            operator: str = None,
            forward: int = 0,
    ) -> tuple[int, int]:
        """
        https://smshub.org/ru/info#getNumbers
        :return: id, number
        """
        params = {
            'action': 'getNumber',
            'service': service,
            'forward': forward
        }
        if country:
            params['country'] = country
        if operator:
            params['operator'] = operator
        response = await self._request("GET", self.BASE_API_URL, params=params)

        if response.text == "NO_NUMBERS":
            raise SmsServiceError(response.text)

        _, id, number = response.text.split(':')
        return int(id), number

    async def _set_status(self, id: str, status: int) -> str:
        """
        https://smshub.org/ru/info#setStatus

        Сразу после получения номера доступны следующие действия:
        8 - Отменить активацию
        1 - Сообщить, что SMS отправлена (необязательно)
        Для активации со статусом 1:
        8 - Отменить активацию
        Сразу после получения кода:
        3 - Запросить еще одну смс
        6 - Подтвердить SMS-код и завершить активацию
        Для активации со статусом 3:
        6 - Подтвердить SMS-код и завершить активацию

        Выходные данные
        Ответ сервера	Описание
        ACCESS_READY	Готовность ожидания смс
        ACCESS_RETRY_GET	Ожидаем новое смс
        ACCESS_ACTIVATION	Активация успешно завершена
        ACCESS_CANCEL	Активация отменена
        """
        params = {
            'action': 'setStatus',
            'id': id,
            'status': status
        }
        response = await self._request("GET", self.BASE_API_URL, params=params)
        return response.text

    async def request_another_sms(self, id):
        await self._set_status(id, 3)

    async def confirm_activation(self, id):
        await self._set_status(id, 6)

    async def cancel_activation(self, id):
        await self._set_status(id, 8)

    async def request_status(self, id: str) -> str:
        """https://smshub.org/ru/info#getStatus"""
        params = {
            'action': 'getStatus',
            'id': id
        }
        response = await self._request("GET", self.BASE_API_URL, params=params)
        return response.text

    async def request_list_of_countries_and_operators(self):
        params = {
            'cat': 'scripts',
            'act': 'manageActivations',
            'asc': 'getListOfCountriesAndOperators'
        }
        response = await self._request("POST", self.UNOFFICIAL_API_URL, params=params)

        return response.json()

    async def wait_for_code(self, id: int, delay: int = 10, max_wait_time: int = 300) -> str:
        await self._set_status(id, 1)  # Сообщаю о том, что сообщение отправлено
        start_time = time.time()  # Засекаем начальное время

        while True:
            response = await self.request_status(id)
            print(response)
            if response.startswith("STATUS_OK"):
                status, code = response.split(":")
                await self.confirm_activation(id)  # Подтверждаю получение кода
                return code

            if (time.time() - start_time) > max_wait_time:
                # Время ожидания превысило максимально допустимое, прерываем цикл
                raise TimeoutError("Max wait time exceeded")

            await asyncio.sleep(delay)

    async def request_prices(self, service: str, country: str = None) -> dict:
        params = {'action': 'getPrices', 'service': service}
        if country:
            params['country'] = country
        response = await self._request("GET", self.BASE_API_URL, params=params)
        return response.json()

    async def set_max_price(self, service: str, max_price: float, country: str) -> str:
        params = {
            'action': 'setMaxPrice',
            'service': service,
            'maxPrice': max_price,
            'country': country
        }
        response = await self._request("GET", self.BASE_API_URL, params=params)
        return response.text

    async def request_current_activations(self) -> dict:
        params = {'action': 'getCurrentActivations'}
        response = await self._request("GET", self.BASE_API_URL, params=params)
        return response.json()



from .client import TwitterClient
from .account import TwitterAccount, TwitterAccountStatus
from .utils import remove_at_sign, parse_oauth_html, tweet_url
from . import errors

__all__ = [
    "TwitterClient",
    "TwitterAccount",
    "TwitterAccountStatus",
    "remove_at_sign",
    "parse_oauth_html",
    "tweet_url",
    "errors",
]



def to_invite_code(invite_code_or_url: str) -> str:
    """
    Extracts the Discord invite code from a URL or a direct code.

    :param invite_code_or_url: Invite code or URL.
    :return: Invite code.
    """
    return invite_code_or_url.split('/')[-1]
